!********************************************
! denproj.f90 - density projection subroutine
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

subroutine density_proj
  implicit none
  include 'globalpm.fh'
  include 'mpif.h'

  character (len=MSL) :: fn,fp 
  integer(kind=4) :: i,isec,jsec,fu
  real(kind=4) :: redshift,z_write
  real(kind=8) :: rhof_tot,rhof_node
  integer(kind=4) :: ierr
  integer(kind=4) :: thrdnum,omp_get_thread_num  
  external omp_get_thread_num

  rho_pxy=0.0
  rho_pxz=0.0
  rho_pyz=0.0
  rhof_node=0.0
  rhof_tot=0.0

  if (master) then
    redshift=1.0/a-1.0
    z_write=z_rp(cur_rp)
  endif
  call mpi_bcast(redshift,1,mpi_real,0,mpi_comm_world,ierr)
  call mpi_bcast(z_write,1,mpi_real,0,mpi_comm_world,ierr)
 
! pass particles
  call particle_pass

! rebuild linked list
  call link_list
  
! make projections
  !$omp parallel default(shared) &
  !$omp private(jsec,isec,thrdnum) &
  !$omp reduction(+:rhof_node)
  thrdnum=omp_get_thread_num()+1
  !$omp do schedule (static,1)
  do jsec=1,NN
     do isec=1,NN
       call rho_build(isec,jsec,thrdnum,rhof_node)
     enddo
  enddo
  !$omp end do
  !$omp end parallel

  call mpi_reduce(rhof_node,rhof_tot,1,mpi_double_precision,mpi_sum,0, &
         mpi_comm_world,ierr)
  if (master) write(*,*) 'average overdensity=',rhof_tot

! delete buffer particles
  nploc=nploc-ngloc

! write files
  do i=0,2
    if (i.eq.0) write (fn,'(f7.3,"xy",i1,".proj")') z_write, mr
    if (i.eq.1) write (fn,'(f7.3,"xz",i1,".proj")') z_write, mr
    if (i.eq.2) write (fn,'(f7.3,"yz",i1,".proj")') z_write, mr
    fn=adjustl(fn)
    fp=PROJPATH//fn
    fu=10+i
    open(unit=fu,file=fp,form='binary') 
    write(fu) redshift
    if (i.eq.0) write(fu) rho_pxy
    if (i.eq.1) write(fu) rho_pxz
    if (i.eq.2) write(fu) rho_pyz
    close(fu)
  enddo

!update projection counter
  cur_rp=cur_rp+1

end subroutine density_proj

subroutine rho_build(isec,jsec,thrdnum,rhof_node)
  implicit none
  include 'globalpm.fh'

  integer(kind=4) :: isec,jsec,thrdnum
  real(kind=8) :: rhof_node
  integer(kind=4) :: ip,i,j,k,i2,j2,jlcic,jhcic,ihcic,ilcic
  integer(kind=4) :: rxl,ryl

  rhof(:LF,:,:,thrdnum)=-1.0
  rhof(LF+1:,:,:,thrdnum)=0.0

  ilcic=LPSSC*(isec-1)+1-LCB
  ihcic=LPSSC*isec+LCB
  jlcic=LPSSC*(jsec-1)+1-LCB
  jhcic=LPSSC*jsec+LCB

  do k=HOCZL,HOCZH
    do j=jlcic,jhcic
      j2=mod(j+LC-1,LC)+1
      do i=ilcic,ihcic
        i2=mod(i+LC-1,LC)+1
        ip=hoc(i2,j2,k,1)
        call cic_mass(ip,i,j,isec,jsec,thrdnum)
      enddo
    enddo
  enddo

  rxl=(isec-1)*LPSS
  ryl=(jsec-1)*LPSS

  do k=1,LPSS
    do j=1,LPSS
      do i=1,LPSS
        rho_pxy(rxl+i,ryl+j)=rho_pxy(rxl+i,ryl+j)+ &
                             rhof(LFB+i,LFB+j,LFB+k,thrdnum)
        rho_pxz(rxl+i,k)=rho_pxz(rxl+i,k)+ &
                         rhof(LFB+i,LFB+j,LFB+k,thrdnum)
        rho_pyz(ryl+j,k)=rho_pyz(ryl+j,k)+ &
                         rhof(LFB+i,LFB+j,LFB+k,thrdnum)
        rhof_node=rhof_node+real(rhof(LFB+i,LFB+j,LFB+k,thrdnum),kind=8)
      enddo
    enddo
  enddo

end subroutine rho_build
