!**********************************************************
! ffftw.f90 - subroutines for use of filefftw subprogram
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!**********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

!**********************************************************
! The filefftw subprogram calculates fftw transforms
! using a different set of processes than those which run
! pmfast.  Data is best shared through a tmpfs ramdisk.

  subroutine filefftw_init
    implicit none

    include 'globalpm.fh'
    character(len=MSL) :: fn 

    if (mr.eq.0) then
      fn=FFTWSWAP//'pm2.stat'
      open(20,file=fn,action='write',form='unformatted',status='replace')
      write(unit=20) 0 
      close(20)
    endif

  end subroutine filefftw_init

!*********************************************************************

  subroutine filefftw_finalize
    implicit none

    include 'globalpm.fh'
    character(len=MSL) :: fn 

    if (mr.eq.0) then
      fn=FFTWSWAP//'pm2.stat'
      open(20,file=fn,action='write',form='unformatted',status='replace')
      write(unit=20) 3 
      close(20)
    endif

  end subroutine filefftw_finalize

!*********************************************************************

  subroutine filefftw(n,data,dir_in)

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4), intent(in) :: n
    integer(kind=4), intent(in) :: dir_in
    real(kind=4), intent(inout) :: data(n+2,n,n/NN)

    integer(kind=4) :: dir,comp,i,j,k,fftwerr
    character(len=MSL) :: fn,fns

    if (dir_in.eq.-1) then
      dir=2
    else
      dir=dir_in
    endif

    comp=0

! --- Distribute Data to files ---

    do i=1,NCPUPN
      write(fns,'("pm2.dat",i1)') i
      fn=FFTWSWAP//fns
      open(50+i,file=fn,action='write',form='unformatted',status='replace')
      !write(*,*) 'fft',i,1+(i-1)*n/NN/NCPUPN,i*n/NN/NCPUPN
      write(50+i) data(1:n+2,1:n,1+(i-1)*n/NN/NCPUPN:i*n/NN/NCPUPN)
      close(50+i)
    enddo

    call mpi_barrier(MPI_COMM_WORLD,fftwerr)

! --- Lock file handshake ---

! First initialize the fft

    if (mr.eq.0) then
#ifdef DIAG
      write(*,*) 'initializing fftw'
#endif
      fn=FFTWSWAP//'pm2.stat'
      open(20,file=fn,action='write',form='unformatted', status='replace')
      write(unit=20) dir
      close(20)

! Loop waiting for a reply

      fn=FFTWSWAP//'fftw.stat'
      do while (comp.ne.2)
        call system('sleep 1')
        open(23,file=fn,action='read',status='old',err=40,form='unformatted')
        read(unit=23,err=40,end=40) comp
40      close(23)
      enddo

! Give the go ahead

      dir=3
      fn=FFTWSWAP//'pm2.stat'
      open(20,file=fn,action='write',form='unformatted',status='replace')
      write(unit=20) dir
      close(20)

! Loop for completion

      fn=FFTWSWAP//'fftw.stat'
      do while (comp.ne.1)
        call system('sleep 1')
        open(23,file=fn,form='unformatted',action='read',status='old',err=41)
        read(unit=23,err=41,end=41) comp
41      close(23)
      enddo
#ifdef DIAG
      write(*,*) 'received completion'
#endif
    endif
    call mpi_barrier(MPI_COMM_WORLD,fftwerr)

! --- read in the transformed data ---

    do i=1,NCPUPN
      write(fns,'("fftw.dat",i1)') i
      fn=FFTWSWAP//fns
      open(60+i,file=fn,action='read',form='unformatted',status='old')
      read(60+i) data(1:n+2,1:n,1+(i-1)*n/NN/NCPUPN:i*n/NN/NCPUPN)
      close(60+i)
    enddo

! --- send back completion of fftw ---

    dir=0
    if (mr.eq.0) then
      fn=FFTWSWAP//'pm2.stat'
      open(24,file=fn,action='write',form='unformatted',status='replace')
      write(unit=24) dir
      close(24)
      fn=FFTWSWAP//'fftw.stat'
      do while (comp.ne.0)
        call system('sleep 1')
        open(25,file=fn, form='unformatted',action='read',status='old',err=42)
        read(unit=25,err=42,end=42) comp
42      close(25)
      enddo
    endif
    call mpi_barrier(MPI_COMM_WORLD,fftwerr)

  end subroutine filefftw
