! filefftw.f90 - interface to fftw mpi functionality through a file
!                exchange mechanism
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

program filefft
  implicit none

  include 'mpif.h'
#ifdef IA64
  character(len=*), parameter :: fpath='/mnt/ramdisk/filefftw/'
#else
  character(len=*), parameter :: fpath='/scratch/merz/'
#endif
  integer(kind=4), parameter :: n2=DEFN2
  integer(kind=4), parameter :: nn=DEFNN 
  integer(kind=4), parameter :: ncpu=DEFNCPU
  integer(kind=4), parameter :: tot_ncpu=nn*ncpu
  
  integer(kind=4) :: istop,ierr,mr,fdir,comp,fdir1
  real(kind=4), dimension(n2+2,n2,n2/tot_ncpu)  :: fdata
  integer(kind=4) :: i,j,k
  integer(kind=8) :: plan, iplan
  character(len=80) :: fn_temp,fft_stat,pm2_stat,fft_dat,pm2_dat

  call mpistart 

  call fftwconf(1)

  fft_stat=fpath//'fftw.stat'
  pm2_stat=fpath//'pm2.stat'
  fn_temp=''
  write(fn_temp,'(a7,i1)') 'pm2.dat',mod(mr,ncpu)+1
  pm2_dat=fpath//fn_temp
  fn_temp=''
  write(fn_temp,'(a8,i1)') 'fftw.dat',mod(mr,ncpu)+1 
  fft_dat=fpath//fn_temp 
 
  fdir=0
  fdir1=1
  comp=0 

  if (mr.eq.0) then
    open(23,file=fft_stat,action='write', &
         status='replace',form='unformatted')
    write(unit=23) comp
    close(23)
    write(*,*) 'initialized, waiting for data'
  endif

  do while(fdir.eq.0) 
    call system('sleep 1')
    if (mr.eq.0) then
      open(20,file=pm2_stat, &
            action='read',status='old',err=41, form='unformatted')
      read(unit=20,err=41,end=41) fdir
41    close(20)
    endif
    call mpi_bcast(fdir,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)

    if (fdir.eq.1.or.fdir.eq.2) then

      comp=2
      if (mr.eq.0) then
        write(*,*) 'fft initiated'
        open(23,file=fft_stat,action='write', &
            status='replace',form='unformatted')
        write(unit=23) comp
        close(23)
        do while (fdir1.ne.3)
          call system('sleep 1')
          open(20,file=pm2_stat,  &
                action='read',status='old',err=42,form='unformatted')
          read(unit=20,err=42,end=42) fdir1
42        close(20)
        enddo
      endif
      call mpi_barrier(MPI_COMM_WORLD,ierr)

      !write(*,*) 'mr:',mr,'starting to read in data'
      open(21,file=pm2_dat,action='read', &
            status='old',form='unformatted')
      !write(*,*) 'fft mr',mr,n2/tot_ncpu
      read(unit=21) fdata(1:n2+2,1:n2,1:n2/tot_ncpu)
      close(21)

      !write(*,*) 'mr:',mr,'starting fftw'

      call mpi_barrier(mpi_comm_world,ierr)
      call mpifft(fdata,tot_ncpu,fdir)

      !write(*,*) 'mr:',mr,'writing out data'

      open(22,file=fft_dat,action='write',  &
            form='unformatted',status='replace')
      write(unit=22) fdata(1:n2+2,1:n2,1:n2/tot_ncpu)
      close(22)

      call mpi_barrier(mpi_comm_world,ierr)

      comp=1
      if (mr.eq.0) then
        open(24,file=fft_stat,action='write',  &
              form='unformatted',status='replace')
        write(unit=24) comp
        close(24)
        write(*,*) 'finished fft'
        do while (fdir1.ne.0)
          call system('sleep 1')
          open(20,file=pm2_stat,  &
                action='read',status='old',err=43,form='unformatted')
          read(unit=20,err=43,end=43) fdir1
43        close(20)
        enddo
        write(*,*) 'resetting'
        comp=0
        open(25,file=fft_stat,action='write',  &
              form='unformatted',status='replace')
        write(unit=25) comp
        close(25)
      endif
      fdir=0
    endif   
  enddo

  write(*,*)'mr:',mr,'exiting file fftw program'

  call fftwconf(0)

  call mpi_finalize(ierr)

contains  
  
!! ---------------------------------------------------------
!! Initialization for message passing and calculation of
!! workload distribution parameters

  subroutine mpistart

    implicit none

    integer(kind=4) :: nn_ret,err_code

    err_code=-1
    call mpi_init(ierr)
    if (ierr.ne.mpi_success) call mpi_abort(MPI_COMM_WORLD,err_code,ierr)
    call mpi_comm_size(mpi_comm_world, nn_ret, ierr)
    if (ierr.ne.mpi_success) call mpi_abort(MPI_COMM_WORLD,err_code,ierr)
    call mpi_comm_rank(mpi_comm_world, mr, ierr)
    if (ierr.ne.mpi_success) call mpi_abort(MPI_COMM_WORLD,err_code,ierr)

    if (nn_ret.ne.tot_ncpu) then
      write(*,*) 'program compiled for different number of total cpus', &
                 nn_ret,'!=',tot_ncpu
      ierr=-1
      call mpi_abort(MPI_COMM_WORLD,err_code,ierr)
   endif

    if (mr.eq.0) then
      write(*,*) 'File IO fftw code initialized'
      write(*,*) 'Array length = ',n2,' over ',nn_ret,' total cpus'
    endif

    return
  end subroutine mpistart

!! --------------------------------------------------------------
!! fftw configuration subroutine

  subroutine fftwconf(inout)
    implicit none

    include 'fftw_f77.i'
    integer(kind=4) :: inout


    if (inout.eq.1) then
      call rfftw3d_f77_mpi_create_plan(plan,MPI_COMM_WORLD,n2, &
            n2,n2, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE)  
      call rfftw3d_f77_mpi_create_plan(iplan,MPI_COMM_WORLD,n2, &
            n2,n2, FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE)
    else
       call rfftwnd_f77_mpi_destroy_plan(iplan)
       call rfftwnd_f77_mpi_destroy_plan(plan)
    endif

    return
  end subroutine fftwconf
  

!! --------------------------------------------------------------
!! This routine must be compiled with :
!!     -lrfftw_mpi -lrfftw -lfftw_mpi -lfftw -lm

  subroutine mpifft(data,ncpu_tot,dir)

    implicit none

    include 'fftw_f77.i'

    integer ncpu_tot,dir,order
    real data(2*(n2/2+1),n2,n2/ncpu_tot), work(2*(n2/2+1),n2,n2/ncpu_tot)

    order=FFTW_NORMAL_ORDER

!! We can use either FFTW_NORMAL_ORDER or FFTW_TRANSPOSED_ORDER

!! Forward direction

    if (dir.eq.1) then
       call rfftwnd_f77_mpi(plan,1,data,work,1,order)

!! Reverse

    elseif (dir.eq.2) then
       call rfftwnd_f77_mpi(iplan,1,data,work,1,order)

!! the following line is an optional normalization

       data(:,:,:) = data(:,:,:)/real(n2*n2*n2)
    else
       if (mr.eq.0) write (*,*) 'FFT direction invalid, dir=',dir
       call MPI_Abort(MPI_COMM_WORLD,ierr)
    endif

    return
  end subroutine mpifft

!! ------------------------------------------------

end program filefft  
