!***********************************
! finepm.f90 - fine mesh subroutines 
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  subroutine finepm(isec,jsec,thrdnum,fdentemp) 

    implicit none
    include 'globalpm.fh'  
 
    integer(kind=4) :: isec,jsec,thrdnum
    real(kind=8), intent(inout) :: fdentemp
    
    integer(kind=4) :: ip,i,j,k,icl,ich,icp,icn,ics, &
                       i2,j2,k2,k1,ilcic,ihcic,jlcic,jhcic, &
                       iref,jref

!! Normalize rho to calculate overdensity

    rhof(:LF,:,:,thrdnum)=-1.0
    rhof(LF+1:,:,:,thrdnum)=0.0

!! Calculate offsets for cube within coarse grid hoc

    ilcic=LPSSC*(isec-1)+1-LCB
    ihcic=LPSSC*isec+LCB
    jlcic=LPSSC*(jsec-1)+1-LCB
    jhcic=LPSSC*jsec+LCB

!! Assign density to fine grid

    do k=HOCZL,HOCZH
      do j=jlcic,jhcic
        j2=mod(j+LC-1,LC)+1
        do i=ilcic,ihcic
          i2=mod(i+LC-1,LC)+1
          ip=hoc(i2,j2,k,1)
          call cic_mass(ip,i,j,isec,jsec,thrdnum)
        enddo
      enddo
    enddo

!! Assign density to coarse grid

    if (coarse_step) call density_bin(isec,jsec,thrdnum)

#ifdef DIAG
    do k=1+LFB,LF-LFB
      do j=1+LFB,LF-LFB
        do i=1+LFB,LF-LFB
          fdentemp=fdentemp+real(rhof(i,j,k,thrdnum),kind=8)
        enddo
      enddo
    enddo
#endif

!! Transform fine grid density to Fourier space

    call fft3dipp(rhof(1,1,1,thrdnum), &
                  rhoft(1,1,1,thrdnum),LF,LF,LF,1)

!! Perform convolution to calculate Fourier space potential

    do k=1,LF
      do j=1,LF/2+1
        do i=1,2*LF,2
          i2=(i-1)/2+1
          rhoft(i,j,k,thrdnum)=rhoft(i,j,k,thrdnum)*fgk(i2,j,k)
          rhoft(i+1,j,k,thrdnum)=rhoft(i+1,j,k,thrdnum)*fgk(i2,j,k)
        enddo
      enddo
    enddo

!! Transform potential back to real space

    call fft3dipp(rhof(1,1,1,thrdnum), &
                  rhoft(1,1,1,thrdnum),LF,LF,LF,-1)

!! Update particle velocities using fine grid force

    call fine_velocity(isec,jsec,thrdnum)

  end subroutine finepm

!*******

  subroutine cic_mass(ip,icell,jcell,isec,jsec,thrdnum)

    implicit none
    include 'globalpm.fh'

    integer(kind=4) :: icell,jcell,isec,jsec,thrdnum
    integer(kind=4), intent(inout) :: ip
    integer(kind=4) :: i1,i2,j1,j2,k1,k2,iref,jref,kref
    real(kind=4) :: pp_x,pp_y,x,y,z,dx1,dx2,dy1,dy2,dz1,dz2

    iref=-(isec-1)*LPSS+LFB
    jref=-(jsec-1)*LPSS+LFB
    kref=LFB

    if (icell.lt.1) then
      pp_x=-0.5-RLPS+iref
    elseif (icell.gt.LC) then
      pp_x=-0.5+RLPS+iref
    else
      pp_x=-0.5+iref
    endif

    if (jcell.lt.1) then
      pp_y=-0.5-RLPS+jref
    elseif (jcell.gt.LC) then
      pp_y=-0.5+RLPS+jref
    else
      pp_y=-0.5+jref
    endif

    do
      if (ip.eq.0) exit
      x=xvp(1,ip)+pp_x
      y=xvp(2,ip)+pp_y
      z=xvp(3,ip)-0.5+kref

      i1=floor(x)+1
      i2=i1+1
      dx1=i1-x
      dx2=1-dx1
      j1=floor(y)+1
      j2=j1+1
      dy1=j1-y
      dy2=1-dy1
      k1=floor(z)+1
      k2=k1+1
      dz1=k1-z
      dz2=1-dz1

      dx1=MP*dx1
      dx2=MP*dx2

      rhof(i1,j1,k1,thrdnum)=rhof(i1,j1,k1,thrdnum)+dx1*dy1*dz1
      rhof(i2,j1,k1,thrdnum)=rhof(i2,j1,k1,thrdnum)+dx2*dy1*dz1
      rhof(i1,j2,k1,thrdnum)=rhof(i1,j2,k1,thrdnum)+dx1*dy2*dz1
      rhof(i2,j2,k1,thrdnum)=rhof(i2,j2,k1,thrdnum)+dx2*dy2*dz1
      rhof(i1,j1,k2,thrdnum)=rhof(i1,j1,k2,thrdnum)+dx1*dy1*dz2
      rhof(i2,j1,k2,thrdnum)=rhof(i2,j1,k2,thrdnum)+dx2*dy1*dz2
      rhof(i1,j2,k2,thrdnum)=rhof(i1,j2,k2,thrdnum)+dx1*dy2*dz2
      rhof(i2,j2,k2,thrdnum)=rhof(i2,j2,k2,thrdnum)+dx2*dy2*dz2

      ip=ll(ip)
    enddo

  end subroutine cic_mass

!*******

  subroutine cic_mass_coarse(ip)

    implicit none
    include 'globalpm.fh'

    integer(kind=4), intent(inout) :: ip
    real(kind=4), parameter :: RLC=LC
    integer(kind=4) :: i1,i2,j1,j2,k1,k2
    real(kind=4) :: x,y,z,dx1,dx2,dy1,dy2,dz1,dz2

    do
      if (ip.eq.0) exit
      x=mod(0.25*xvp(1,ip)-0.5+RLC,RLC)
      y=mod(0.25*xvp(2,ip)-0.5+RLC,RLC)
      z=(0.25*xvp(3,ip))-0.5

      i1=floor(x)+1
      i2=mod(i1,LC)+1
      dx1=i1-x
      dx2=1-dx1
      j1=floor(y)+1
      j2=mod(j1,LC)+1
      dy1=j1-y
      dy2=1-dy1
      k1=floor(z)+1
      k2=k1+1
      dz1=k1-z
      dz2=1-dz1

      dx1=MP*dx1
      dx2=MP*dx2

      if (k1.ge.1.and.k1.le.LPSSC) then
        rhoc(i1,j1,k1)=rhoc(i1,j1,k1)+dx1*dy1*dz1
        rhoc(i2,j1,k1)=rhoc(i2,j1,k1)+dx2*dy1*dz1
        rhoc(i1,j2,k1)=rhoc(i1,j2,k1)+dx1*dy2*dz1
        rhoc(i2,j2,k1)=rhoc(i2,j2,k1)+dx2*dy2*dz1
      endif
      if (k2.ge.1.and.k2.le.LPSSC) then
        rhoc(i1,j1,k2)=rhoc(i1,j1,k2)+dx1*dy1*dz2
        rhoc(i2,j1,k2)=rhoc(i2,j1,k2)+dx2*dy1*dz2
        rhoc(i1,j2,k2)=rhoc(i1,j2,k2)+dx1*dy2*dz2
        rhoc(i2,j2,k2)=rhoc(i2,j2,k2)+dx2*dy2*dz2
      endif

      ip=ll(ip)
    enddo

  end subroutine cic_mass_coarse

!*******

  subroutine density_bin(isec,jsec,thrdnum)
 
    implicit none
    include 'globalpm.fh'

    integer(kind=4) :: isec,jsec,thrdnum
    integer(kind=4) :: i,j,k,i1,j1,k1
    integer(kind=4) :: ip

    do k=0,LPSSC+1
      do j=(jsec-1)*LPSSC+1,jsec*LPSSC 
        do i=(isec-1)*LPSSC+1,isec*LPSSC 
          ip=hoc(i,j,k,1)
          call cic_mass_coarse(ip)
        enddo
      enddo
    enddo

  end subroutine density_bin

!*******

  subroutine fine_velocity(isec,jsec,thrdnum)

    implicit none
    include 'globalpm.fh'

    integer (kind=4) :: isec,jsec,thrdnum
    integer(kind=4) :: i,j,k,ip,km,kp,jm,jp,im,iu,ii,i1,i2,j1,j2,k1,k2
    real(kind=4) :: x,y,z,dx1,dx2,dy1,dy2,dz1,dz2,dV,x_off,y_off,z_off
    real(kind=4) :: force_mag 

    force_f(:,:,:,:,thrdnum)=0.0
 
    do k=LFB-1,LF-LFB+1
      km=k-1
      kp=k+1
      do j=LFB-1,LF-LFB+1
        jm=j-1
        jp=j+1
        do i=LFB-1,LF-LFB+1
          im=i-1
          iu=i+1
          force_f(1,i,j,k,thrdnum)=(rhof(im,j,k,thrdnum) &
                            -rhof(iu,j,k,thrdnum))*0.5
          force_f(2,i,j,k,thrdnum)=(rhof(i,jm,k,thrdnum) &
                            -rhof(i,jp,k,thrdnum))*0.5
          force_f(3,i,j,k,thrdnum)=(rhof(i,j,km,thrdnum) &
                            -rhof(i,j,kp,thrdnum))*0.5
        enddo
      enddo
    enddo

    force_mag=0.0
    do k=LFB-1,LF-LFB+1
      do j=LFB-1,LF-LFB+1
        do i=LFB-1,LF-LFB+1
          force_mag=force_f(1,i,j,k,thrdnum)*force_f(1,i,j,k,thrdnum) &
               +force_f(2,i,j,k,thrdnum)*force_f(2,i,j,k,thrdnum) &
               +force_f(3,i,j,k,thrdnum)*force_f(3,i,j,k,thrdnum)
          force_mag=sqrt(force_mag) 
          if (force_mag.gt.fmax(thrdnum)) then
            fmax(thrdnum)=force_mag
            fmaxloc(:,thrdnum)=(/i+(isec-1)*LPSS, &
                                   j+(jsec-1)*LPSS,k+mr*LPSS/)
          endif
        enddo
      enddo
    enddo

    x_off=-0.5+real(LFB)-(isec-1)*LPSS
    y_off=-0.5+real(LFB)-(jsec-1)*LPSS
    z_off=-0.5+real(LFB)

    do k=1,LPSSC
      do j=(jsec-1)*LPSSC+1,jsec*LPSSC
        do i=(isec-1)*LPSSC+1,isec*LPSSC
          ip=hoc(i,j,k,1)
          do
            if (ip.eq.0) exit
           
            x=xvp(1,ip)+x_off
            y=xvp(2,ip)+y_off
            z=xvp(3,ip)+z_off
            i1=floor(x)+1
            i2=i1+1
            dx1=i1-x
            dx2=1-dx1
            j1=floor(y)+1
            j2=j1+1
            dy1=j1-y
            dy2=1-dy1
            k1=floor(z)+1
            k2=k1+1
            dz1=k1-z
            dz2=1-dz1

            dV=aGdt*dx1*dy1*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i1,j1,k1,thrdnum)*dV
            dV=aGdt*dx2*dy1*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i2,j1,k1,thrdnum)*dV
            dV=aGdt*dx1*dy2*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i1,j2,k1,thrdnum)*dV
            dV=aGdt*dx2*dy2*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i2,j2,k1,thrdnum)*dV
            dV=aGdt*dx1*dy1*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i1,j1,k2,thrdnum)*dV
            dV=aGdt*dx2*dy1*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i2,j1,k2,thrdnum)*dV
            dV=aGdt*dx1*dy2*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i1,j2,k2,thrdnum)*dV
            dV=aGdt*dx2*dy2*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_f(1:3,i2,j2,k2,thrdnum)*dV

            ip=ll(ip)
          enddo
        enddo
      enddo
    enddo

end subroutine fine_velocity
