#include <stdio.h> 
#include <stdlib.h> 
#include <time.h> 
#include <sys/types.h> 
#include <math.h> 
//#include <ia64intrin.h>
#include "ipp.h" 
#include "ippm.h"
#include "ipps.h"

#define NBUF 100000

//static int gdummy=0;

static IppsFFTSpec_C_64fc* pFFTSpeccc;
static int ncc=0;
  
static IppsFFTSpec_R_64f* pFFTSpecrc;
static int nrc=0;

static IppsFFTSpec_C_32fc* pFFTSpecscc;
static int nscc=0;
  
static IppsFFTSpec_R_32f* pFFTSpecsrc;
static int nsrc=0;

/*  the initialization routine is not re-entrant and should not be called
    from within openmp loops
*/
fft1dippccinit_(int *np)
{
  int pow2,order;

  if (*np == ncc) return;
  if (*np != 0) ippsFFTFree_C_64fc(pFFTSpeccc); 
  ncc=*np;
  if (ncc > NBUF/10) printf("fft1dippcc: NBUF may be too small\n");
  pow2=ncc;
  order=log((double) ncc)/log(2.)+0.5; 
  if (((int)pow(2,order)+0.5) != ncc) 
		printf("fft1dippcc: not power of 2: n=%d\n",ncc);
  ippsMalloc_64fc(pow2);  
  ippsFFTInitAlloc_C_64fc(&pFFTSpeccc,order, IPP_FFT_DIV_INV_BY_N,ippAlgHintFast);  
}

fft1dippsccinit_(int *np)
{
  int pow2,order;

  if (*np == nscc) return;
  if (*np != 0) ippsFFTFree_C_32fc(pFFTSpecscc); 
  nscc=*np;
  if (nscc > NBUF/10) printf("fft1dippcc: NBUF may be too small\n");
  pow2=nscc;
  order=log((double) nscc)/log(2.)+0.5; 
  ippsMalloc_32fc(pow2);  
  ippsFFTInitAlloc_C_32fc(&pFFTSpecscc,order, IPP_FFT_DIV_INV_BY_N,ippAlgHintFast);  
}

fft1dipprcinit_(int *np)
{
  int pow2,order;

  if (*np == nrc) return;
  if (*np != 0) ippsFFTFree_R_64f(pFFTSpecrc); 
  nrc=*np;
  if (nrc > NBUF/10) printf("fft1dipprc: NBUF may be too small\n");
  pow2=nrc;
  order=log((double) nrc)/log(2.)+0.5; 
  ippsMalloc_64f(pow2);  
  ippsFFTInitAlloc_R_64f(&pFFTSpecrc,order, IPP_FFT_DIV_INV_BY_N,ippAlgHintFast);  
}
  
fft1dippsrcinit_(int *np)
{
  int pow2,order;

  if (*np == nsrc) return;
  if (*np != 0) ippsFFTFree_R_32f(pFFTSpecsrc); 
  nsrc=*np;
  if (nsrc > NBUF/10) printf("fft1dipprc: NBUF may be too small\n");
  pow2=nsrc;
  order=log((double) nsrc)/log(2.)+0.5; 
  ippsMalloc_32f(pow2);  
  ippsFFTInitAlloc_R_32f(&pFFTSpecsrc,order, IPP_FFT_DIV_INV_BY_N,ippAlgHintFast);  
}
  
#  define asm_prefetch(p,len) \
          __asm__ __volatile__( \
                " \n\
				shr r2=%1,6   ;;		 \n\t\
				add r2=-1,r2   ;;		 \n\t\
				mov ar.lc=r2   	;;		\n\t\
				add r2=64,%0 ;;		\n\t\
		.LF1:\n\t\
                                lfetch.nta         [%0],64 \n\t \
			        br.cloop.sptk.few  .LF1     \n\t \ 
				;;		\n\t\
                 \n" \
                    :: "r" (p), "r"(len):"ar.lc","r2" \
                     );


typedef char prefetcht;
void prefetch(prefetcht *a, int len)
{
	int i,j=0;
	if (len<64) return;
#ifdef	__ia64__
	asm_prefetch(a+len,len);
//	for (i=0;i<len;i+=64) __lfetch(__lfhint_nta,a+i);
#endif
//	gdummy=j;
}


// the 1d fft Complex to Complex using IPP library
// parameters:
//             a - the input complex array of size n
//             b - the output complex array of size n
//             np - the dimensions of arrays
//             isignp - the direction of the transform
//                      1 - forward
//                      -1 - backward
int fft1dippcc_(Ipp64fc *a, Ipp64fc *b, int *isignp){
  
  Ipp8u pWork[NBUF];
 
   //fft transformation:
     
   if (*isignp==1){
     ippsFFTFwd_CToC_64fc(a, b, pFFTSpeccc, pWork);
   } else {
     ippsFFTInv_CToC_64fc(a, b, pFFTSpeccc, pWork);
   } 
}


int fft1dippscc_(Ipp32fc *a, Ipp32fc *b, int *isignp){
  
  Ipp8u pWork[NBUF];
 
//   ippsAddC_32f_I(1.e-30,a,2*nscc);
   //fft transformation:
   prefetch(a,2*nsrc*sizeof(Ipp32f));
   if (a != b) prefetch(b,2*nsrc*sizeof(Ipp32f));
   if (*isignp==1){
     ippsFFTFwd_CToC_32fc(a, b, pFFTSpecscc, pWork);
   } else {
     ippsFFTInv_CToC_32fc(a, b, pFFTSpecscc, pWork);
   } 
}


// the 1d fft Real to Complex (and back) using IPP library
//  parameters:
//              a - the input real array of size n
//              b - the output complex array of size n
//              np - the dimensions of arrays
//              isignp - the direction of the transform
//                       1 - forward
//                       -1 - backward
int fft1dipprc_(Ipp64f *a, Ipp64f *b, int *isignp){
    
  Ipp8u pWork[NBUF];
    
   //fft transformation 
   if (*isignp==1){
     ippsFFTFwd_RToCCS_64f(a, b, pFFTSpecrc, pWork);
   } else {
     ippsFFTInv_CCSToR_64f(a, b, pFFTSpecrc, pWork);
   }
 
}


int fft1dippsrc_(Ipp32f *a, Ipp32f *b, int *isignp){
    
   Ipp8u pWork[NBUF];
   IppStatus status;
    
//   status=ippsAddC_32f_I(1.e-30,a,nsrc);
//   if(status != 0) printf("fft1dippsrc error=%d\n",status);

   prefetch(a,nsrc*sizeof(Ipp32f));
   if (a != b) prefetch(b,2*nsrc*sizeof(Ipp32f));

   //fft transformation 
   if (*isignp==1){
     status=ippsFFTFwd_RToCCS_32f(a, b, pFFTSpecsrc, pWork);
   } else {
     status=ippsFFTInv_CCSToR_32f(a, b, pFFTSpecsrc, pWork);
   }
   if (status != 0) {
	printf("fft1dippsrc error=%d isign=%d message=%s\n",status,*isignp,
		ippGetStatusString(status));
        printf("a=%ld b=%ld pFFTSpecsrc=%d\n",a,b,pFFTSpecsrc);
        printf("nsrc=%d\n",nsrc);
   }
}
int ippscale_(Ipp64f *a, Ipp64f *val, int *np)
{
    int status;
    if (status=ippsMulC_64f_I(*val,a,*np) != 0) {
	printf("ippscale error status=%d\n",status);
    }
}

int ippcopy8_(Ipp64f *dest, Ipp64f *src, int *np)
{
    int status;
    if (status=ippsCopy_64f(src,dest,*np) != 0) {
	printf("ippscale error status=%d\n",status);
    }
}

int ippscales_(Ipp32f *a, Ipp32f *val, int *np)
{
    int status;
    if (status=ippsMulC_32f_I(*val,a,*np) != 0) {
	printf("ippscale error status=%d\n",status);
    }
}


// tranpose single precision real matrix by faking it as complex.
int ipptransposes_(Ipp64f *dest, Ipp64f *src, int *nxp, int *nyp)
{
  int status;
  Ipp32u width,height;

  width=*nyp;
  height=*nxp;
//  printf("transpose: nx=%d ny=%d\n",width,height);
  if (status=ippmTranspose_m_64f(src,width*sizeof(Ipp64f),width,height,dest,height*sizeof(Ipp64f)) != 0){
    printf("ipptransposes error status=%d\n",status);
  }
}

int prefetch_(float *f, int *len)
{
	int i;
//	for (i=0;i<*len;i+=8) __builtin_prefetch(f+i,1,3);
}
int t6x6_(Ipp64f *dest, Ipp64f *src,int *nxp, int *nyp)
{
   ippmTranspose_m_64f_6x6(src,*nyp*sizeof(Ipp64f),dest,*nxp*sizeof(Ipp64f));
}
