!*********************************************
! kernel.f90 - kernel construction subroutines
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


!******************************************************************
!! fine grid potential kernel construction

  subroutine fgkernel

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: ierr,errcode

    integer(kind=4) :: i,j,k,temp(3)
    real(kind=4) :: x,y,z,r

    rhof(:,:,:,:)=0.0

!! Copy kernel into volume

    if (lsf_kern) then
      open(unit=18,file=KERNPATH//'kernelfineall.dat')
    else
      open(unit=18,file=KERNPATH//'wg1.ascii')
    endif
    do k=1,SRC
      do j=1,SRC
        do i=1,SRC
          read(18,*) temp(1),temp(2),temp(3),rhof(i,j,k,1)
          if (temp(1).ne.i.or.temp(2).ne.j.or.temp(3).ne.k) then
            write(*,*) 'error reading in fine mesh kernel'
            errcode=-1
            call mpi_abort(mpi_comm_world,errcode,ierr)
          endif
        end do
      end do
    end do
    close(18)

!! Reflect accross y/2 plane 

      do j=2,SRC
        rhof(1:SRC,LF-j+2,1:SRC,1)=rhof(1:SRC,j,1:SRC,1)
      enddo

!! Reflect accross x/2 plane
          
      do i=2,SRC
        rhof(LF-i+2,1:LF,1:SRC,1)=rhof(i,1:LF,1:SRC,1)
      enddo

!! Reflect accross z/2 plane

      do k=2,SRC
        rhof(1:LF,1:LF,LF-k+2,1)=rhof(1:LF,1:LF,k,1)
      enddo

!! Transform to Fourier space 

      call fft3dipp(rhof(1,1,1,1),rhoft(1,1,1,1),LF,LF,LF,1)

!! Select real portion for potential kernel

      !$omp parallel do default(shared) private(i,j,k)
      do k=1,LF
        do j=1,LF/2+1
          do i=1,LF
            fgk(i,j,k)=rhoft(2*i-1,j,k,1)
          enddo
        enddo
      enddo
      !$omp end parallel do

    end subroutine fgkernel

!****************************************************************
!! coarse grid force kernel construction 

    subroutine cgkernel
   
      implicit none
      include 'mpif.h'
      include 'globalpm.fh'
    
      integer(4) :: ierr
      real(4), parameter :: pi = 3.14159625
      integer(kind=4) :: i,j,k,HC1,temp(3)
      integer(kind=8) :: xvp_mem,ck_mem
      real(kind=4) :: x,y,z,r,rmax
      character(len=MSL) :: kfile,kfilepath
      integer(4) :: kr,ki,kx,ky,kz
      real(4) :: ka,kb,kc,wa,wb,wc

      rmax=4*LC
      HC1=LC/2+1
      ck=0.0
      force_c=0.0

!! Construct force kernel
   
      do k=1,HC1
        z=4.0*(k-1)
        do j=1,HC1
          y=4.0*(j-1)
          do i=1,HC1
            x=4.0*(i-1)
            r=sqrt(x*x+y*y+z*z)
            if (r.eq.0.0) then
              ck(:,i,j,k)=0.0
            else
              ck(1,i,j,k)=-x/r**3
              ck(2,i,j,k)=-y/r**3
              ck(3,i,j,k)=-z/r**3
            endif
          enddo
        enddo
      enddo

!! Read in kernel for 2-level matching

      if (lsf_kern) then
        open(unit=11,file=KERNPATH//'kernelcoarseall.dat')
        do k=1,LCB+1 
          do j=1,LCB+1 
            do i=1,LCB+1 
              read(11,*) temp(:),ck(:,i,j,k)
            enddo
          enddo
        enddo
        close(11)
      else
        open(unit=11,file=KERNPATH//'wfxyz2.ascii')
        do k=1,LCB-1 
          do j=1,LCB-1 
            do i=1,LCB-1 
              read(11,*) temp(:),ck(:,i,j,k)
            enddo
          enddo
        enddo
        close(11)
      endif

!! Reflect in y plane

      do j=2,LC/2
        ck(1,1:HC1,LC-j+2,1:HC1)=ck(1,1:HC1,j,1:HC1)
        ck(2,1:HC1,LC-j+2,1:HC1)=-ck(2,1:HC1,j,1:HC1)
        ck(3,1:HC1,LC-j+2,1:HC1)=ck(3,1:HC1,j,1:HC1)
      enddo

!! Reflect in x plane

      do i=2,LC/2
        ck(1,LC-i+2,1:LC,1:HC1)=-ck(1,i,1:LC,1:HC1)
        ck(2:3,LC-i+2,1:LC,1:HC1)=ck(2:3,i,1:LC,1:HC1)
      enddo

!! Reflect in z plane

      do k=2,LC/2
        ck(1:2,1:LC,1:LC,LC-k+2)=ck(1:2,1:LC,1:LC,k)
        ck(3,1:LC,1:LC,LC-k+2)=-ck(3,1:LC,1:LC,k)
      enddo

!! Copy local nodes portion into kernel

      force_c(:,:,:,1:LPSSC)=ck(:,:,:,1+mr*LPSSC:(mr+1)*LPSSC)

#ifdef LRCKCORR

!! Overwrite ck with non-corrected kernels

!! Construct force kernel

      do k=1,HC1
        z=4.0*(k-1)
        do j=1,HC1
          y=4.0*(j-1)
          do i=1,HC1
            x=4.0*(i-1)
            r=sqrt(x*x+y*y+z*z)
            if (r.eq.0.0) then
              ck(:,i,j,k)=0.0
            else
              ck(1,i,j,k)=-x/r**3
              ck(2,i,j,k)=-y/r**3
              ck(3,i,j,k)=-z/r**3
            endif
          enddo
        enddo
      enddo

!! Reflect in y plane

      do j=2,LC/2
        ck(1,1:HC1,LC-j+2,1:HC1)=ck(1,1:HC1,j,1:HC1)
        ck(2,1:HC1,LC-j+2,1:HC1)=-ck(2,1:HC1,j,1:HC1)
        ck(3,1:HC1,LC-j+2,1:HC1)=ck(3,1:HC1,j,1:HC1)
      enddo

!! Reflect in x plane

      do i=2,LC/2
        ck(1,LC-i+2,1:LC,1:HC1)=-ck(1,i,1:LC,1:HC1)
        ck(2:3,LC-i+2,1:LC,1:HC1)=ck(2:3,i,1:LC,1:HC1)
      enddo

!! Reflect in z plane

      do k=2,LC/2
        ck(1:2,1:LC,1:LC,LC-k+2)=ck(1:2,1:LC,1:LC,k)
        ck(3,1:LC,1:LC,LC-k+2)=-ck(3,1:LC,1:LC,k)
      enddo

      do k=1,LPSSC
        do j=1,LC
          do i=1,LC+2
            temp_force(i,j,k)=ck(1,i,j,k+mr*LPSSC)
          enddo
        enddo
      enddo
      call filefftw(LC,temp_force,1)
      do k=1,LPSSC
        do j=1,LC
          do i=1,LC+2
            ck(1,i,j,k)=temp_force(i,j,k)
          enddo
        enddo
      enddo

      do k=1,LPSSC
        do j=1,LC
          do i=1,LC+2
            temp_force(i,j,k)=ck(2,i,j,k+mr*LPSSC)
          enddo
        enddo
      enddo
      call filefftw(LC,temp_force,1)
      do k=1,LPSSC
        do j=1,LC
          do i=1,LC+2
            ck(2,i,j,k)=temp_force(i,j,k)
          enddo
        enddo
      enddo

      do k=1,LPSSC
        do j=1,LC
          do i=1,LC+2
            temp_force(i,j,k)=ck(3,i,j,k+mr*LPSSC)
          enddo
        enddo
      enddo
      call filefftw(LC,temp_force,1)
      do k=1,LPSSC
        do j=1,LC
          do i=1,LC+2
            ck(3,i,j,k)=temp_force(i,j,k)
          enddo
        enddo
      enddo

#endif


!! Transform kernel to fourier space

      temp_force(:,:,:)=force_c(1,:,:,1:LPSSC)
      call filefftw(LC,temp_force,1)
      force_c(1,:,:,1:LPSSC)=temp_force(:,:,:)

      temp_force(:,:,:)=force_c(2,:,:,1:LPSSC)
      call filefftw(LC,temp_force,1)
      force_c(2,:,:,1:LPSSC)=temp_force(:,:,:)

      temp_force(:,:,:)=force_c(3,:,:,1:LPSSC)
      call filefftw(LC,temp_force,1)
      force_c(3,:,:,1:LPSSC)=temp_force(:,:,:)

#ifdef DEBUG_KERN

      open(156,file=KERNPATH//'fourier_cgk.dat',form='binary')
      write(156) ck(:,:,:,1:LPSSC) 
!      write(156) force_c(:,:,:,1:LPSSC)
!      do k=1,LPSSC
!        do j=1,LC
!          do i=1,LC/2+1
!            write(156,'(3f20.10)') ck(1,i,j,k),ck(2,i,j,k),ck(3,i,j,k)
!          enddo
!        enddo
!      enddo
      close(156)
      call mpi_barrier(mpi_comm_world,ierr)
      write(*,*) 'rank',mr,'finished writing kernel file'

#endif

#ifdef LRCKCORR

!! Apply long-range corrections

      !$omp parallel do default(shared) private(i,j,k,kr,kx,ky,kz,ka,kb,kc,ki,wa,wb,wc)
      do k=1,LPSSC
        ki=k+LPSSC*mr
        if (ki .lt. LC/2+2) then
           kz=ki-1
        else
           kz=ki-1-LC
        endif
        do j=1,LC
          if (j .lt. LC/2+2) then
             ky=j-1
          else
             ky=j-1-LC
          endif
          do i=1,LC+2,2
             kx=(i-1)/2
             kr = sqrt( real(kx*kx + ky*ky + kz*kz) ) !wtf
             if (kr .le. 8) then
               ka=2*sin(pi*kx/real(LC))
               kb=2*sin(pi*ky/real(LC))
               kc=2*sin(pi*kz/real(LC))
               if (kx .ne. 0) then
                 wa=force_c(1,i+1,j,k)
                 wb=ck(1,i+1,j,k)
                 wc=4*pi*ka/(ka**2+kb**2+kc**2)/16
                 force_c(1,i+1,j,k)=wa*(wc/wb)
               endif
               if (ky .ne. 0) then
                 wa=force_c(2,i+1,j,k)
                 wb=ck(2,i+1,j,k)
                 wc=4*pi*kb/(ka**2+kb**2+kc**2)/16
                 force_c(2,i+1,j,k)=wa*(wc/wb)
               endif
               if (kz .ne. 0) then
                 wa=force_c(3,i+1,j,k)
                 wb=ck(3,i+1,j,k)
                 wc=4*pi*kc/(ka**2+kb**2+kc**2)/16
                 force_c(3,i+1,j,k)=wa*(wc/wb)
               endif
             endif
          enddo
        enddo
      enddo
      !$omp end parallel do

#endif

!! Odd force kernels are purely imaginary
!! Store imaginary quadrant (including Nyquist freq) in cgk

      !$omp parallel do default(shared) private(i,j,k)
      do k=1,LPSSC
        do j=1,LC
          do i=1,LC/2+1
            cgk(:,i,j,k)=force_c(:,2*i,j,k) 
          enddo
        enddo
      enddo
      !$omp end parallel do

#ifdef DEBUG_KERN

!      open(156,file=KERNPATH//'raw_kern.dat',form='binary')
      open(156,file=KERNPATH//'corr_kern.dat',form='binary')
      write(156) cgk
!      do k=1,LPSSC
!        do j=1,LC
!          do i=1,LC/2+1
!            write(156,'(3f20.10)') cgk(1,i,j,k),cgk(2,i,j,k),cgk(3,i,j,k)
!          enddo
!        enddo
!      enddo
      close(156)

#endif

  end subroutine cgkernel

