!***********************************************************
! pairs.f90 - generate and measure pairwise forces
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  subroutine init_pair_vars
    implicit none
    include 'globalpm.fh'
 
      a=1.0
      t=0.0
      dt=1.0
      dtC=1.0
      dt_old=1.0
      nts=nts+1
      aGdt=1.0
      aGdtC=1.0
      xvp=0.0
      sbuf=0.0
      rbuf=0.0
      rhof=0.0
      rhoft=0.0
      rhoc=0.0
      force_c=0.0
      force_f=0.0
      ll=0
      taglist=0
      hoc=0
      toc=0
      nploc=0
      ngloc=0
      coarse_step=.true.

  end subroutine

!*******
  subroutine initialize_pairtest
    implicit none
    include 'mpif.h'
    include 'globalpm.fh'
    integer(kind=4) :: ierr,nn_returned,exit_err

      call mpi_init(ierr)
      call mpi_comm_size(mpi_comm_world,nn_returned,ierr)
      exit_err=-1
      if (nn_returned.ne.NN) then
        write(*,*) 'mismatch in NN compiled / running'
        call mpi_abort(mpi_comm_world,exit_err,ierr)
      endif
      call mpi_comm_rank(mpi_comm_world,mr,ierr)
      if (mr.eq.0) master=.true.
      if (3_8*int(LC,kind=8)**3.gt.6_8*int(NP,kind=8)) then
        write(*,*) 'not enough memory for coarse kernel construction'
        call mpi_abort(mpi_comm_world,exit_err,ierr)
      endif  
      call omp_set_num_threads(NT)
      if (master) then
        call system_clock(count=wcts)
      endif
      call filefftw_init
      if (master) write(*,*) &
           'finished initializing filefftw:', t_elapsed(wcts)
       call fgkernel
       if (master) write(*,*) &
           'finished calculating fine gravity kernel:', t_elapsed(wcts)
       call cgkernel
       if (master) write(*,*) &
           'finished calculating coarse gravity kernel:', t_elapsed(wcts)
       if (master) write(*,*) 'finished initialization:', t_elapsed(wcts) 

  end subroutine initialize_pairtest

!*******

  subroutine pairtest_mesh
    implicit none
    include 'globalpm.fh'
    include 'mpif.h'

    integer(kind=4) :: i,j,k,ip,in,ipp
    integer(kind=4) :: isec,jsec
    real(kind=8) :: fdentemp,fdentot
    real(kind=4) :: dtmacc,dtmmpi

!pairtest variables
    integer(kind=4) :: tag,nprecloc
    integer(kind=4), dimension(MPI_STATUS_SIZE) :: status
    real(kind=4) :: r(3),v(3),rmag,vmag,xdotmag
    real(kind=4), dimension(6,2) :: pair
    real(kind=4), dimension(3,2) :: xdot
    character(len=MSL) :: fname
!end pairtest variables

    integer(kind=4) :: thrdnum,ierr
    integer(kind=4) :: omp_get_thread_num, omp_get_num_threads
    external omp_get_thread_num, omp_get_num_threads
  
    call link_list
    call particle_pass
    call link_list
 
    fdentemp=0.0
    fdentot=0.0
    rhoc(:,:,:)=0.0
    if (nploc.gt.0) write(*,*) 'before fine:', xvp(:,1)
    thrdnum=1
    do jsec=1,NN
       do isec=1,NN
         if (fine) call finepm(isec,jsec,thrdnum,fdentemp,dtmacc)
       enddo
    enddo
    call mpi_barrier(mpi_comm_world,ierr)

    do i=1,nploc
      write(*,*) mr,'after fine:', xvp(:,i)
    enddo
    write(*,*) mr,'ngloc=',ngloc
   
    tag=3421
    nprecloc=nploc-ngloc
    write(*,*) mr,'nprecloc=',nprecloc,'nploc=',nploc
    if (master) then
      if (nprecloc.eq.2) then
        pair(:,:)=xvp(:,1:2)
      elseif (nprecloc.eq.1) then
        pair(:,1)=xvp(:,1)
        call mpi_recv(pair(1:6,2),6,MPI_REAL,MPI_ANY_SOURCE, &
                  tag,MPI_COMM_WORLD,status,ierr)
      else
        call mpi_recv(pair(1:6,1),6,MPI_REAL,MPI_ANY_SOURCE, &
                  tag,MPI_COMM_WORLD,status,ierr)
        call mpi_recv(pair(1:6,2),6,MPI_REAL,MPI_ANY_SOURCE, &
                  tag,MPI_COMM_WORLD,status,ierr)
      endif
    else
      if (nprecloc.eq.1) then
        pair(:,1)=xvp(:,1)
        pair(3,1)=pair(3,1)+mr*LPSS
        call mpi_send(pair(1:6,1),6,mpi_real,0,tag,MPI_COMM_WORLD,ierr)
      elseif (nprecloc.eq.2) then
        pair(:,:)=xvp(:,1:2)
        pair(3,:)=pair(3,:)+mr*LPSS
        call mpi_send(pair(1:6,1),6,mpi_real,0,tag,MPI_COMM_WORLD,ierr)
        call mpi_send(pair(1:6,2),6,mpi_real,0,tag,MPI_COMM_WORLD,ierr)
      endif
    endif
    call mpi_barrier(mpi_comm_world,ierr)
    write(*,*) 'finished sending back'
    if (master) then
      fname=PAIRPATH//'fine_pair.dat'
      open(unit=62,file=fname,position='append')
      write(62,'(12f20.10)') pair(1,1),pair(2,1),pair(3,1),pair(4,1), &
                             pair(5,1),pair(6,1),pair(1,2),pair(2,2), &
                             pair(3,2),pair(4,2),pair(5,2),pair(6,2)
      close(62)
    endif

    call coarsepm 

    nploc=nploc-ngloc

  end subroutine pairtest_mesh

!*******

  subroutine set_pair
    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: i,j,isize
    real(kind=4) :: rnum
    integer(kind=4), dimension(:), allocatable :: iseed
    integer(kind=4) :: ierr
    real(kind=4), dimension(3,2) :: pair
    
    pair=0.0
    
    if (master) then
      write(*,*) 'set pair, nts=',nts
      if (nts.eq.1) then
        pair(1:3,1)=(/74.66510010,84.94259644,10.64086914/)
        pair(1:3,2)=(/81.85389709,76.94387817,159.19999695/)  
        call random_seed(size=isize)
        allocate(iseed(isize))
        do j=1,isize
          iseed(j)=(j+421)*int(t_elapsed(wcts)*10.0,kind=4)
        enddo
        call random_seed(put=iseed(1:isize))
        deallocate(iseed) 
        do i=1,100000
          call random_number(rnum)
        enddo
      else
        j=1              
        do i=1,3
          call random_number(rnum)
          if (rnum.ge.0.9999) rnum=0.995
          pair(i,j)=rnum*LPS
        enddo
        j=2
        do i=1,3
          call random_number(rnum)
          if (rnum.ge.0.9999) rnum=0.995
          pair(i,j)=mod(pair(i,1)+(rnum-0.5)*3.0*(mod(nts,14)+0.2)+  &
                    real(LPS,kind=4),real(LPS,kind=4))
        enddo
      endif
    endif
      
    if (master) write(*,*) 'before scatter'
      
    call mpi_bcast(pair(1:3,1:2),6,mpi_real,0,MPI_COMM_WORLD,ierr)
 
    nploc=0
    ngloc=0
    
    do j=1,2
      if (pair(3,j).ge.mr*LPSS.and.pair(3,j).lt.(mr+1)*LPSS) then
        nploc=nploc+1
        xvp(1:2,nploc)=pair(1:2,j)
        xvp(3,nploc)=pair(3,j)-mr*LPSS
      endif
    enddo
    do i=1,nploc
      write(*,*) 'mr=',mr,'particle',i,'position=',xvp(1:3,i) 
    enddo
    if (master) write(*,*) 'finished pair set'
  
  end subroutine set_pair

!*******
  
  subroutine measure_pair
    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: i,j
    integer(kind=4) :: ierr,tag
    integer(kind=4), dimension(MPI_STATUS_SIZE) :: status
    real(kind=4) :: r(3),v(3),rmag,vmag,xdotmag
    real(kind=4), dimension(6,2) :: pair
    real(kind=4), dimension(3,2) :: xdot
    character(len=MSL) :: fname   
 
    tag=3142

    if (master) then
      write(*,*) 'starting total pair writing'
      if (nploc.eq.2) then
        pair(:,:)=xvp(:,1:2)
      elseif (nploc.eq.1) then
        pair(:,1)=xvp(:,1)
        call mpi_recv(pair(1:6,2),6,MPI_REAL,MPI_ANY_SOURCE, &
                    tag,MPI_COMM_WORLD,status,ierr)
      else
        call mpi_recv(pair(1:6,1),6,MPI_REAL,MPI_ANY_SOURCE, &
                    tag,MPI_COMM_WORLD,status,ierr)
        call mpi_recv(pair(1:6,2),6,MPI_REAL,MPI_ANY_SOURCE, &
                    tag,MPI_COMM_WORLD,status,ierr)
      endif
    else
      if (nploc.eq.1) then
        pair(:,1)=xvp(:,1)
        pair(3,1)=pair(3,1)+mr*LPSS
        call mpi_send(pair(1:6,1),6,mpi_real,0,tag,MPI_COMM_WORLD,ierr)
      elseif (nploc.eq.2) then
        pair(:,:)=xvp(:,1:2)
        pair(3,:)=pair(3,:)+mr*LPSS
        call mpi_send(pair(1:6,1),6,mpi_real,0,tag,MPI_COMM_WORLD,ierr)
        call mpi_send(pair(1:6,2),6,mpi_real,0,tag,MPI_COMM_WORLD,ierr)
      endif
    endif

    if (master) then  
      write(*,*) 'writing out pairs' 
      fname=PAIRPATH//'total_pair.dat' 
      open(unit=73,file=fname,position='append')
      if (nts.eq.1) write(73,'(2i8)') LPS,MAX_NTS
      if (nts.eq.1) write(73,'(2f20.10)') MP,dt 
      write(73,'(12f20.10)') pair(1,1),pair(2,1),pair(3,1),pair(4,1),pair(5,1),  &
               pair(6,1),pair(1,2),pair(2,2),pair(3,2),pair(4,2),pair(5,2),pair(6,2)
      close(73)
    endif
    
    if (master) write(*,*) 'finished pair measure'
    
  end subroutine measure_pair


