!*******************************************
! partpass.f90 - particle passing subroutine
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  subroutine particle_pass
    use qsorti_mod
    implicit none

    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: ip,i,j,k,m,n,nplocr,npt2,npt3
    integer(kind=4), dimension(2) :: npu,npd
    integer(kind=4), dimension(2) :: ngu,ngd
    real(kind=4), dimension(6) :: xvptemp
    real(kind=4) :: zmax,zmin,zgmax,zgmin,ztemp,epsilon
    real(kind=4) :: t_passing    
    integer(kind=4) :: status(MPI_STATUS_SIZE)
    integer(kind=4) :: tag,ierr,err_code,sr,rr
#ifdef NBPASS
    integer(4) :: srequest,rrequest,sierr,rierr
    integer(4), dimension(mpi_status_size) :: sstatus,rstatus
#endif
#ifdef DEBUG
    integer(4) :: num_ghosts_in_list
    integer(4) :: num_real_in_buffer
    integer(4) :: particle_flux
    integer(4) :: num_particles_next_ts
#endif

#ifdef DIAG   
    if (master) then 
      t_passing=0.0
      write(*,*) 'start of passing:',t_elapsed(t_passing)
    endif
#endif

#ifdef DEBUG
   do i=0,NN-1
     if (mr.eq.i) then
       write(*,*) 'mr',mr,'xvp(672181278)',xvp(:,672181278)
       write(*,*) 'mr',mr,'xvp(690012403)',xvp(:,690012403)
     endif
     call mpi_barrier(mpi_comm_world,ierr)
   enddo
#endif

!!  This is where we need to delete the particles that were out of 
!!  bounds during the last link list construction
!!  add to taglist, increment ngloc

    ngloc=0
    do k=1,nt
      do j=1,ndel(k)
        ngloc=ngloc+1
        taglist(ngloc)=dellist(j,k)
      enddo
    enddo
 
    err_code=-1
    zmax=RLPSS-LCB*4
    zmin=real(LCB*4)
    zgmax=RLPSS
    zgmin=0.0
    epsilon=1.0e-04
    npu(:)=0
    npd(:)=0
    ngu(:)=0
    ngd(:)=0

    if (restart_step) then
      m=HOCZL
    else
      m=1
    endif

    do k=m,1+LPPSB
      do j=1,LC
        do i=1,LC
          ip=hoc(i,j,k,1)
          do while (ip.ne.0)
            if (xvp(3,ip).lt.zmin) then
              npd(1)=npd(1)+1
              sbuf((npd(1)-1)*6+1:npd(1)*6)=xvp(:,ip)
              if (xvp(3,ip).lt.zgmin) then
                ngloc=ngloc+1
                taglist(ngloc)=ip
              endif
            endif
            ip=ll(ip)
          enddo
        enddo
      enddo
    enddo

    ngd(1)=ngloc-sum(ndel)
    ngd(2)=ngd(1)
    npd(2)=npd(1)

#ifdef DIAG
    if (master) then
       write(*,*) 'finished downwards buffer:',t_elapsed(t_passing)
       t_passing=0.0
    endif

    if (master) write(*,*) 'downward particle pass'
    do i=0,NN-1
      if (mr.eq.i) then
        write(*,*) 'mr:',mr,'npd:',npd(1),'ngd:',ngd(1)
      endif
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif

    tag = 12
    sr=mod(mr+NN-1,NN)
    rr=mod(mr+1,NN)
    call mpi_sendrecv_replace(npd(2),1, MPI_INTEGER, &
       sr, tag, rr, tag, MPI_COMM_WORLD, status, ierr )
    tag = 13
    call mpi_sendrecv_replace(ngd(2),1, MPI_INTEGER, &
       sr, tag, rr, tag, MPI_COMM_WORLD, status, ierr )

!! Send Down

    !$omp parallel do default(shared) private(i)
    do i=3,6*npd(1),6
      if(sbuf(i).lt.zgmin) then
         sbuf(i)=min(sbuf(i),zgmin-epsilon)+RLPSS
      else
         sbuf(i)=sbuf(i)+RLPSS
         if(sbuf(i).ge.RLPSS+zmin) sbuf(i)=(RLPSS+zmin)-epsilon
      endif
    enddo
    !$omp end parallel do

#ifdef DIAG
    if (master) then
       write(*,*) 'finished downwards boundry check:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

    j=npd(2)*6
    m=npd(1)*6

    if (m.gt.MAX_BUFFER) then
      write(*,*) 'mr:',mr,'exceeded buffer in PARTPASS'
      write(*,*) 'm:',m,'j:',j,'maxbuf:',MAX_BUFFER
      call mpi_abort(mpi_comm_world,err_code,ierr)
    endif

!! could implement non-blocking send recieves followed
!! by a wait - this could even be done using a NN*NN matrix
!! to make sure that particles end up where they belong 
!! regardless of velocity.

#ifdef NBPASS
    call mpi_isend(sbuf,m,mpi_real,sr,sr, &
                   mpi_comm_world, srequest, sierr)
    call mpi_irecv(rbuf,j,mpi_real,rr,mr, &
                   mpi_comm_world, rrequest, rierr)
    call mpi_wait(srequest,sstatus,sierr)
    call mpi_wait(rrequest,rstatus,rierr)
#else
    if (mod(mr,2).eq.0) then
      call mpi_send(sbuf,m,MPI_REAL, &
                    sr,tag,MPI_COMM_WORLD,ierr)
    else
      call mpi_recv(rbuf,j,MPI_REAL, &
                    rr,tag,MPI_COMM_WORLD,status,ierr)
    endif

    call mpi_barrier(mpi_comm_world,ierr)

    if (mod(mr,2).eq.1) then
      call mpi_send(sbuf,m,MPI_REAL, &
                    sr,tag,MPI_COMM_WORLD,ierr)
    else
      call mpi_recv(rbuf,j,MPI_REAL, &
                    rr,tag,MPI_COMM_WORLD,status,ierr)
    endif
#endif

    call mpi_barrier(mpi_comm_world,ierr)

#ifdef DIAG
    if (master) then
       write(*,*) 'finished downwards pass:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

    !$omp parallel do default(shared) private(i,m)
    do i=1,npd(2)
      m=(i-1)*6+1
      rbuf(m)=modulo(rbuf(m),RLPS)
      rbuf(m+1)=modulo(rbuf(m+1),RLPS)
      xvp(:,nploc+i)=rbuf(m:m+5)
    enddo
    !$omp end parallel do

#ifdef DIAG
    if (master) then
       write(*,*) 'finished unpacking downwards:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

!! Start of upwards pass
!! Would require same alterations as downward pass

    if (restart_step) then
      restart_step=.false.
      m=HOCZH
    else
      m=LPSSC
    endif

    do k=LPSSC-LPPSB,m
      do j=1,LC
        do i=1,LC
          ip=hoc(i,j,k,1)
          do while (ip.ne.0)
            if (xvp(3,ip).ge.zmax) then
              npu(1)=npu(1)+1
              sbuf((npu(1)-1)*6+1:npu(1)*6)=xvp(:,ip)
              if (xvp(3,ip).ge.zgmax) then
                ngloc=ngloc+1
                taglist(ngloc)=ip
              endif
            endif
            ip=ll(ip)
          enddo
        enddo
      enddo
    enddo

    ngu(1)=ngloc-ngd(1)
    ngu(2)=ngu(1)
    npu(2)=npu(1)

#ifdef DIAG
    if (master) then
       write(*,*) 'finished upwards buffer:',t_elapsed(t_passing)
       t_passing=0.0
    endif
    if (master) write(*,*) 'upward particle pass'
    do i=0,NN-1
      if (mr.eq.i) then
        write(*,*) 'mr:',mr,'npu:',npu(1),'ngu:',ngu(1)
      endif
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif

    tag = 11
    sr=mod(mr+1,NN)
    rr=mod(mr+NN-1,NN)
    call mpi_sendrecv_replace(npu(2),1,MPI_INTEGER, &
               sr,tag,rr,tag,MPI_COMM_WORLD,status,ierr )
    tag = 15
    call mpi_sendrecv_replace(ngu(2),1,MPI_INTEGER, &
               sr,tag,rr,tag,MPI_COMM_WORLD,status,ierr )

!! Send Up

   !$omp parallel do default(shared) private(i)
   do i=3,6*npu(1),6
     if (sbuf(i).ge.zgmax) then
       sbuf(i)=max(sbuf(i),zgmax+epsilon)-RLPSS
     else
       sbuf(i)=sbuf(i)-RLPSS
       if (sbuf(i).lt.zmax-RLPSS) sbuf(i)=(zmax-RLPSS)+epsilon
     endif
   enddo
   !$omp end parallel do

#ifdef DIAG
    if (master) then
       write(*,*) 'finished upwards boundry check:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

    j=npu(2)*6
    m=npu(1)*6

    if (m.gt.MAX_BUFFER) then
      write(*,*) 'mr:',mr,'exceeded buffer in PARTPASS'
      write(*,*) 'm:',m,'j:',j,'maxbuf:',MAX_BUFFER
      call mpi_abort(mpi_comm_world,err_code,ierr)
    endif

!! Same case as above

#ifdef NBPASS
    call mpi_isend(sbuf,m,mpi_real,sr,sr, &
                   mpi_comm_world, srequest, sierr)
    call mpi_irecv(rbuf,j,mpi_real,rr,mr, &
                   mpi_comm_world, rrequest, rierr)
    call mpi_wait(srequest,sstatus,sierr)
    call mpi_wait(rrequest,rstatus,rierr)
#else
    if (mod(mr,2).eq.0) then
      call mpi_send(sbuf,m,MPI_REAL, &
                    sr,tag,MPI_COMM_WORLD,ierr)
    else
      call mpi_recv(rbuf,j,MPI_REAL, &
                    rr,tag,MPI_COMM_WORLD,status,ierr)
    endif

    call mpi_barrier(mpi_comm_world,ierr)

    if (mod(mr,2).eq.1) then
       call mpi_send(sbuf,m,MPI_REAL, &
                     sr,tag,MPI_COMM_WORLD,ierr)
    else
       call mpi_recv(rbuf,j,MPI_REAL, &
                     rr,tag,MPI_COMM_WORLD,status,ierr)
    endif
#endif

    call mpi_barrier(mpi_comm_world,ierr)

#ifdef DIAG
    if (master) then
       write(*,*) 'finished upwards pass:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

    !$omp parallel do default(shared) private(i,m)
    do i=1,npu(2)
      m=(i-1)*6+1
      rbuf(m)=modulo(rbuf(m),RLPS)
      rbuf(m+1)=modulo(rbuf(m+1),RLPS)
      xvp(:,nploc+npd(2)+i)=rbuf(m:m+5)
    enddo
    !$omp end parallel do

    if (ngloc.gt.MAX_TAG) then
      write(*,*) 'mr:',mr,'exceeded taglist in PARTPASS'
      write(*,*) 'ngloc:',ngloc,'maxtag:',MAX_TAG
      call mpi_abort(mpi_comm_world,err_code,ierr)
    endif

#ifdef DEBUG
    num_ghosts_in_list = 0
    do i=1,nploc
      if (xvp(3,i).lt.zgmin.or.xvp(3,i).ge.zgmax) &
        num_ghosts_in_list = num_ghosts_in_list + 1
    enddo
    num_real_in_buffer = 0
    do i=nploc+1,nploc+npu(2)+npd(2)
      if (xvp(3,i).ge.zgmin.and.xvp(3,i).lt.zgmax) &
        num_real_in_buffer = num_real_in_buffer + 1
    enddo
    particle_flux = num_real_in_buffer - num_ghosts_in_list
    num_particles_next_ts = nploc + particle_flux
    do i=0,NN-1
      if (mr.eq.i) then
        write(*,*) 'NUMBER GHOSTS FOR',mr,'=',num_ghosts_in_list
        write(*,*) 'number found by',mr,'=',ngloc,'out of',nploc
        write(*,*) 'number of particles at next timestep=',num_particles_next_ts
        if (ngloc.ne.num_ghosts_in_list) &
          call mpi_abort(mpi_comm_world,ierr,ierr)     
      endif
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif

#ifdef DIAG
    if (master) then
       write(*,*) 'finished upwards unpacking:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

!! Now we must sort the taglist so we can shuffle ghost particles to
!! the end of the particle list 

    if (ngloc.gt.0) call qsorti(taglist(:ngloc))

#ifdef DIAG
    if (master) then
       write(*,*) 'finished sorting taglist:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

!! Now we can move the ghost particles to the end of the original 
!! particle list

    m=nploc+1
    do i=ngloc,1,-1
      m=m-1
      xvptemp(:)=xvp(:,m)
      xvp(:,m)=xvp(:,taglist(i))
      xvp(:,taglist(i))=xvptemp(:)
    enddo

#ifdef DIAG
    if (master) then
       write(*,*) 'finished moving ghosts:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

#ifdef DEBUG
    do i=1,nploc-ngloc
      if (xvp(3,i).lt.zgmin.or.xvp(3,i).ge.zgmax) then
        write(*,*) 'mr',mr,'ghost not moved to end of list'
        write(*,*) xvp(3,i),i
        call mpi_abort(mpi_comm_world,ierr,ierr)
      endif
    enddo
#endif

!! Loop through buffer and deleted particles
!! swap the real particles arriving from other nodes to the
!! start of the buffer region
!! Start at end, work back.

   nplocr=nploc-ngloc+1              !start ghosts + buffer
   nploc=nploc+npu(2)+npd(2)         !all particles

   npt2=ngu(2)-ngu(1)+ngd(2)-ngd(1)  !expected particle flux

#ifdef DEBUG
   if (npt2.ne.particle_flux) then
     write(*,*) 'mr',mr,'particle flux differs'
     write(*,*) 'expected',npt2,'measured',particle_flux
     call mpi_abort(mpi_comm_world,ierr,ierr)
   endif
#endif 
     
   npt3=nplocr-1+npt2                !number of real particles 

#ifdef DEBUG 
   do i=0,NN-1
     if (mr.eq.i) then
       write(*,*) 'my rank=',mr
       write(*,*) 'start of new real particles=',nplocr
       write(*,*) 'expected particle flux=',npt2
       write(*,*) 'number of real particles=',npt3
     endif
     call mpi_barrier(mpi_comm_world,ierr)
   enddo
  do i=1,nplocr-1
    ztemp=xvp(3,i)
    if (ztemp.lt.zgmin.or.ztemp.ge.zgmax) then
      write(*,*) 'after shuffling back ghosts, ghost in list',mr,ztemp,i
      write(*,*) 'this is incorrect!'
    endif
  enddo
#endif

!! append real particles to the end of the particle list

   ip=nploc        !start at very end
   do 
     if (ip.lt.nplocr) exit   !exit if we process all local particles 
     if (xvp(3,ip).ge.zgmin.and.xvp(3,ip).lt.zgmax) then  !if local
       xvptemp(:)=xvp(:,ip)
       xvp(:,ip)=xvp(:,nplocr)  !swap with particle immediately following list
       xvp(:,nplocr)=xvptemp(:) !append local particle to list
       nplocr=nplocr+1          !extend local particle list
       ip=ip+1                  !don't move forward in buffer, check swappee 
     endif
     ip=ip-1
   enddo

#ifdef DIAG
    if (master) then
       write(*,*) 'finished shuffling reals:',t_elapsed(t_passing)
       t_passing=0.0
    endif
#endif

#ifdef DEBUG 
  do i=0,NN-1
    if (mr.eq.i) then
      write(*,*) mr,'end of real particles stepped=',nplocr-1
    endif
    call mpi_barrier(mpi_comm_world,ierr)
  enddo
  do i=1,nplocr-1
     ztemp=xvp(3,i)     
     if (ztemp.lt.zgmin.or.ztemp.ge.zgmax) then
       write(*,*) 'ghost in real region!',mr,ztemp,i
       call mpi_abort(mpi_comm_world,ierr,ierr)
     endif
   enddo
   do i=nplocr,nploc
     ztemp=xvp(3,i)
     if (ztemp.ge.zgmin.and.ztemp.lt.zgmax) then
       write(*,*) 'real in ghost region!',mr,ztemp,i
       call mpi_abort(mpi_comm_world,ierr,ierr)
     endif
   enddo
#endif

   ngloc=nploc-nplocr+1
      
#ifdef DIAG
    if (master) then
      write(*,*) 'particle passing statistics'
      write(*,*) 'mr nploc ngloc nplocr npu(1:2) npd(1:2) ngu(1:2) ngd(1:2)'
    endif
    do i=0,NN-1
      if (mr.eq.i) then
        write(*,*) mr,nploc,ngloc,nplocr,npu(1),npu(2),npd(1),npd(2),ngu(1),ngu(2),ngd(1),ngd(2)
        write(*,*) xvp(:,nplocr)
        write(*,*) 'expected particle flux=',npt2
      endif
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif

#ifdef DEBUG
   do i=0,NN-1
     if (mr.eq.i) then
       write(*,*) 'mr',mr,'xvp(672181278)',xvp(:,672181278)
       write(*,*) 'mr',mr,'xvp(690012403)',xvp(:,690012403)
     endif
     call mpi_barrier(mpi_comm_world,ierr)
   enddo
#endif

  end subroutine particle_pass
