!********************************************************************
! Initialization subroutine for pm2mpi
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  subroutine initialize_pm

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: nn_returned
    integer(kind=4) :: ip,i,j,k,m
    integer(kind=4) :: isize
    real(kind=4) :: rnum
    integer(kind=4), dimension(:), allocatable :: iseed
    character (len=MSL) :: np_ifile,ifile,ofile
    integer(kind=8) :: np_in
    real(kind=4) :: np_tot_in,z_restart_file
    real(kind=8), dimension(6) :: avg
    logical :: initerr
    integer(kind=4) :: ierr,abort_err
    integer(kind=4) :: mr_restart


    call initialize_common
    call initialize_mpi

!! Check to see if there is enough memory for coarse kernel

   abort_err=-1
   if (3_8*int(LC,kind=8)**3.gt.6_8*int(NP,kind=8)) then
        write(*,*) 'not enough memory for coarse kernel construction'
        call mpi_abort(mpi_comm_world,abort_err,ierr) 
   endif
 
!! Set number of threads to use

    call omp_set_num_threads(NT)

!! Check to make sure COARSE is set properly

    if (COARSE.ne.0.and.mod(COARSE,2).ne.1) then
      if (master) write(*,*) 'COARSE must be either 0 or an odd integer'
      initerr=.true.
    endif

!! Check to make sure we are using proper communications for odd
!! number of nodes

#ifdef NBPASS
    if (master) write(*,*) 'using non-blocking particle passing'
#else
    if (mod(NN,2).ne.0) then
      if (master) write(*,*) 'odd number of nodes, pmfast must be compiled with -DNBPASS'
        call mpi_abort(mpi_comm_world,abort_err,ierr)
    endif
#endif

!! Initialize Timer, Display Memory Usage for common block

    if (master) then
      call system_clock(count=wcts)
      call max_mem
      write(*,*) 'Starting pm2mpi:', t_elapsed(wcts)
    endif

!! Read in redshifts for checkpointing + density projections

  if (master) then
    open(11,file='input/checkpoints')
    do num_cp=1,MAX_NUM_CP
      read(unit=11,err=51,end=41,fmt='(f5.1)') z_cp(num_cp)
    enddo
    41  num_cp=num_cp-1
    51  close(11)
    if (num_cp.eq.MAX_NUM_CP) then
      write(*,*) 'too many checkpoints to store'
      call mpi_abort(mpi_comm_world,abort_err,ierr)
    endif
    do i=1,num_cp
      a_cp(i)=1.0/(1.0+z_cp(i))
    enddo
    write(*,*) 'Checkpointing performed at:'
    write(*,*) 'z      a'
    do i=1,num_cp
      write(*,'(f5.1,2x,f6.4)') z_cp(i),a_cp(i)
    enddo

    open(11,file='input/projections')
    do num_rp=1,MAX_NUM_RP
      read(unit=11,err=61,end=71,fmt='(f5.1)') z_rp(num_rp)
    enddo
    71  num_rp=num_rp-1
    61  close(11)
    if (num_rp.eq.MAX_NUM_RP) then
      write(*,*) 'too many projections to store'
      call mpi_abort(mpi_comm_world,abort_err,ierr)
    endif
    do i=1,num_rp
      a_rp(i)=1.0/(1.0+z_rp(i))
    enddo
    write(*,*) 'Projections performed at:'
    write(*,*) 'z        a'
    do i=1,num_rp
      write(*,'(f7.4,2x,f6.4)') z_rp(i),a_rp(i)
    enddo

  endif

  cur_cp=1
  cur_rp=1

  dt=0.5
  dt_acc=0.5
  a=AI
  aGdt=dt/(6.0*acos(-1.0))

!! initialize fft's and potential kernels

  call filefftw_init

  if (master) write(*,*) &
     'finished initializing filefftw:', t_elapsed(wcts)

  call fgkernel

  if (master) write(*,*) &
     'finished calculating fine gravity kernel:', t_elapsed(wcts)

  call cgkernel

  if (master) write(*,*) &
     'finished calculating coarse gravity kernel:', t_elapsed(wcts)

!! Initialize particles 

  nploc=NP/NN 

    initialize: select case(INIT_VAL)
       
       case(1) !one particle on mr = 0

         if (master) then
           nploc=1
           xvp(1:3,1)=RLPSS/2-0.5+1
!!! 8 particles in a coarse grid cell.
if (.false.) then
           nploc=8
           xvp(1:3,1)=RLPSS/2-0.5+3
           xvp(1:3,2)=RLPSS/2-0.5+2
           xvp(1:3,3)=RLPSS/2-0.5+3
           xvp(1,3)=RLPSS/2-0.5+2
           xvp(1:3,4)=RLPSS/2-0.5+3
           xvp(2,4)=RLPSS/2-0.5+2

           xvp(1:3,5)=RLPSS/2-0.5+2
           xvp(3,5)=RLPSS/2-0.5+3

           xvp(1:3,6)=RLPSS/2-0.5+3
           xvp(3,6)=RLPSS/2-0.5+2

           xvp(1:3,7)=RLPSS/2-0.5+2
           xvp(2,7)=RLPSS/2-0.5+3

           xvp(1:3,8)=RLPSS/2-0.5+2
           xvp(1,8)=RLPSS/2-0.5+3
         

           xvp(4:6,1:8)=0.0
endif
           xvp(4:6,1)=0.0
         else
           nploc=0
         endif

       case(2) !randomly distributed

         nploc=NP/NN
         call random_seed(size=isize)
         allocate(iseed(isize))
         do j=1,isize
           iseed(j)=(mr*isize)+j+341
         enddo
         call random_seed(put=iseed(1:isize))
         do ip=1,nploc
           call random_number(rnum)
           if (rnum.ge.0.9999) then
             rnum=0.995
           endif
           xvp(1,ip)=rnum*RLPS
           call random_number(rnum)
           if (rnum.ge.0.9999) then
             rnum=0.995
           endif
           xvp(2,ip)=rnum*RLPS
           call random_number(rnum)
           if (rnum.ge.0.9999) then
             rnum=0.995
           endif
           xvp(3,ip)=rnum*RLPSS
           xvp(4:6,ip)=rnum - 0.5 !0.0
         enddo
         deallocate(iseed)

       case(3,5:6) !homogeneously distributed (3-D Cubic Grid)

         nploc=NP/NN
         ip = 0
         do k=0,LPSS-1,2
           do j=0,LPS-1,2
             do i=0,LPS-1,2
               ip=ip+1
               xvp(1,ip) = i + 0.5
               xvp(2,ip) = j + 0.5
               xvp(3,ip) = k + 0.5
               if (ip.gt.nploc) then
                 if (master) write(*,*) 'Too Many Particles'
                 initerr=.true.
               endif
             enddo
           enddo
         enddo
         if (INIT_VAL.eq.3) then
           xvp(4:6,:) = 0.0
         elseif (INIT_VAL.eq.5) then
           xvp(4:5,:) = 0.0
           xvp(6,:) =0.25
         elseif (INIT_VAL.eq.6) then
           xvp(4:5,:) = 0.0
           xvp(6,:) =1.0
         endif

       case(4) !homogeneously distributed, with 4 rows displaced in grid

         nploc=NP/NN
         ip=0
         do k=0,LPSS-1,2
           do j=0,LPS-1,2
             do i=0,LPS-1,2
               ip=ip+1
               xvp(1,ip) = i + 0.5
               xvp(2,ip) = j + 0.5
               if (k.gt.LPSS-8) then
                 xvp(3,ip) = k - 5.0  + 0.5
               else
                 xvp(3,ip) = k + 0.5
               endif
               if (ip.gt.nploc) then
                 write(*,*) ip,nploc,i,j,k
                 if (master) write(*,*) 'Too Many Particles'
                 initerr=.true.
               endif
             enddo
           enddo
         enddo

         xvp(4:6,:) = 0.0

       case(7) !two particles
       if (.false.) then
         if (mr.eq.0) then
           nploc=1
           xvp(1:2,1)=LPS/2
           xvp(3,1)=LPSS-8
           xvp(4:6,1)=0.0
         elseif (mr.eq.1) then
           nploc=1
           xvp(1:2,1)=LPS/2
           xvp(3,1)=8
           xvp(4:6,1)=0.0
         else
           nploc=0
         endif
       endif

         if (mr.eq.0) then
           nploc=2
           xvp(1:2,1)=LPS/2
           xvp(3,1:2)=LPSS/2
           xvp(1,2)=LPS/2
           xvp(2,2)=LPS/2+16
           xvp(4:6,:)=0.0
         else
           nploc=0
         endif

       case(8) !read in from files

           write(ofile,'("xvp",i1,".init")') mr
           ifile=INICOND//ofile
       !    ifile=INICOND//'xvp.init'
           open(unit=83,file=ifile,form='binary',status='old')
           read(83) nploc

           do i=0,NN-1
             if (mr.eq.i) then
               write(*,*) 'mr:',mr,'# particles =',nploc
             endif
             call mpi_barrier(MPI_COMM_WORLD,ierr)
           enddo

           if (master) write(*,*) 'number of particles expected =',NP
           call mpi_reduce(real(nploc,kind=4),np_tot_in,1,MPI_REAL, &
                           MPI_SUM,0,MPI_COMM_WORLD,ierr)
           if (master) then 
             write(*,*) 'number of particles total =',int(np_tot_in,kind=8)
             if (int(np_tot_in,kind=8).ne.NP) then
               write(*,*) 'number particles read in incorrect'
               initerr=.true.
             endif
           endif

           read(83) xvp(1:6,1:nploc)
           close(83)

         avg(:)=0.0
         do ip=1,nploc
           if (xvp(3,ip).lt.0.0.or.xvp(3,ip).ge.RLPSS) then
             write(*,*) 'mr',mr,'particle out of z bounds'
             write(*,*) xvp(3,ip),'0.0',RLPSS
           endif
           avg(:)=avg(:)+real(xvp(:,ip),kind=8)
         enddo
         avg(:)=avg(:)/real(nploc,kind=8)
         write(6,'(a3,i3,1x,a15,1x,6f9.3)') 'mr:',mr,'average inital:',avg(:)

       case(9) !restart

         restart_step=.true.

         if (master) then
           write(ifile,'(f5.1,"params.dat")') z_restart
           ifile=adjustl(ifile)
           ofile=OUT1//ifile
           open (unit=14,file=ofile,form='binary')
           read(14) a,t,nts,dt_acc,dt_acc_c,cur_cp,cur_rp
           close(14)
         endif

         call mpi_bcast(cur_cp,1,mpi_integer,0,mpi_comm_world,ierr)
         call mpi_bcast(cur_rp,1,mpi_integer,0,mpi_comm_world,ierr)
 
         write(ifile,'("xvp",i1,".dat")') mr
         if (mod(cur_cp+1,2).eq.1) then
           ofile=OUT1//ifile
         else
           ofile=OUT2//ifile
         endif
         open (unit=12,file=ofile,form='binary')
         read(12) nploc
         read(12) xvp(1:6,1:nploc)
         close(12)

         call initial_pass

       case default !incorrect init_val

         write(*,*) 'please use correct initialization value'
         initerr=.true.

       end select initialize

       if (initerr) call mpi_abort(mpi_comm_world,abort_err,ierr)

       if (master) write(*,*) 'finished initialization:', t_elapsed(wcts)

  end subroutine initialize_pm

!*******

  subroutine initialize_common
    implicit none
    include 'globalpm.fh'

    mr=0
    nploc=0
    ngloc=0
    nts=0
    num_cp=0
    cur_cp=0
    num_rp=0
    cur_rp=0
 
    t=-3.0/sqrt(AI)
    dt=0.0
    dt_old=0.0
    dt_acc=0.0
    dt_acc_c=0.0
    a=0.0
    aGdt=0.0
    wcts=0.0

    xvp=0.0
    sbuf=0.0
    rbuf=0.0
    rhof=0.0
    rhoft=0.0
    rho_pxy=0.0
    rho_pxz=0.0
    rho_pyz=0.0
    rhoc=0.0
    force_c=0.0
    force_f=0.0
    fgk=0.0
    cgk=0.0
    ll=0
    taglist=0
    hoc=0
    toc=0
    
    master=.false.
    stopsim=.false.
    check_step=.false.
    coarse_step=.false.
    restart_step=.false.
    proj_step=.false.

  end subroutine initialize_common

!*********

  subroutine max_mem
    implicit none
    include 'globalpm.fh'

    real(kind=4) :: mem_tot

    mem_tot= (6.0*real(MAX_PARTICLES)) + &  ! xvp
             real(MAX_BUFFER)      + &        ! sbuf
             real(MAX_BUFFER)      + &        ! rbuf
             real(LF+2*LF*LF*NT)   + &        ! rhof
             real(2*LF*(LF/2+1)*LF*NT) + &    ! rhoft
             real((LC+2)*LC*LC/NN)       + &    ! rhoc
             real(3*(LC+2)*(LC)*(LC/NN+2)) + &  ! force_c
             real(LPS*LPS)             + &    ! rho_pxy
             real(LPS*LPSS)            + &    ! rho_pxz
             real(LPS*LPSS)            + &    ! rho_pyz
             real(LC*LC*3)               + &    ! force_buffer
             real((LC+2)*LC*(LC/NN))     + &    ! temp_force
             real(3*(LF-2*LFB)**3*NT)    + &    ! force_f
             real(LF*(LF/2+1)*LF)      + &    ! fgk
             real(3*(LC/2+1)*LC*(LC/NN))    + &    ! cgk
             real(MAX_PARTICLES)   + &    ! ll
             real(MAX_TAG)             + &    ! taglist
             real(LC*LC*(LC/NN+2*LCB)*NT) + & ! hoc
             real(LC*LC*(LC/NN+2*LCB)*NT)     ! toc


    write(*,*) 'majority of memory used / node :'
    write(*,*) mem_tot*4.0,'bytes'
    write(*,*) mem_tot*4.0/1024/1024/1024,'GB'

  end subroutine max_mem

!**************************************************************

  subroutine initialize_mpi
    implicit none
    include 'mpif.h'
    include 'globalpm.fh'
    
    integer(kind=4) :: ierr,nn_returned,exit_err

    call mpi_init(ierr)
    call mpi_comm_size(mpi_comm_world,nn_returned,ierr)
    if (nn_returned.ne.NN) then
      write(*,*) 'mismatch in NN compiled / running'
      exit_err=-1
      call mpi_abort(mpi_comm_world,exit_err,ierr)
    endif
    call mpi_comm_rank(mpi_comm_world,mr,ierr)
    if (mr.eq.0) master=.true. 

  end subroutine initialize_mpi

!*******
  subroutine initial_pass
    use qsorti_mod
    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,m
    integer(kind=4) :: npu(2),npd(2)
    integer(kind=4) :: status(MPI_STATUS_SIZE)
    integer(kind=4) :: tag,ierr,err_code,sr,rr
#ifdef NBPASS
    integer(4) :: srequest,rrequest,sierr,rierr
    integer(4), dimension(mpi_status_size) :: sstatus,rstatus
#endif
    real(kind=4) :: xvptemp(6)

    err_code=-1
    npd(:)=0
    npu(:)=0
    do i=1,nploc
      if (xvp(3,i).lt.0.0) then
        npd(1)=npd(1)+1
        taglist(npd(1))=i
        sbuf((npd(1)-1)*6+1:npd(1)*6)=xvp(:,i)
      endif
    enddo

    if (npd(1).ge.MAX_TAG) then
      write(*,*) mr,'ran out of tag space'
      call mpi_abort(mpi_comm_world,err_code,ierr)
    endif

    npd(2)=npd(1)
    tag = 12
    sr=mod(mr+NN-1,NN)
    rr=mod(mr+1,NN)
    call mpi_sendrecv_replace(npd(2),1, MPI_INTEGER, &
        sr, tag, rr, tag, MPI_COMM_WORLD, status, ierr )

    if (npd(1)*6.ge.MAX_BUFFER) then
      write(*,*) mr,'ran out of buffer space'
      call mpi_abort(mpi_comm_world,err_code,ierr)
    endif

    do i=3,6*npd(1),6
      sbuf(i)=sbuf(i)+RLPSS
      if (sbuf(i).ge.RLPSS) sbuf(i)=RLPSS-1.0e-3
    enddo

    j=npd(2)*6
    m=npd(1)*6


#ifdef NBPASS
    call mpi_isend(sbuf,m,mpi_real,sr,sr, &
                   mpi_comm_world, srequest, sierr)
    call mpi_irecv(rbuf,j,mpi_real,rr,mr, &
                   mpi_comm_world, rrequest, rierr)
    call mpi_wait(srequest,sstatus,sierr)
    call mpi_wait(rrequest,rstatus,rierr)
#else
    if (mod(mr,2).eq.0) then
      call mpi_send(sbuf,m,MPI_REAL, &
                    sr,tag,MPI_COMM_WORLD,ierr)
    else
      call mpi_recv(rbuf,j,MPI_REAL, &
                    rr,tag,MPI_COMM_WORLD,status,ierr)
    endif

    call mpi_barrier(mpi_comm_world,ierr)

    if (mod(mr,2).eq.1) then
       call mpi_send(sbuf,m,MPI_REAL, &
                     sr,tag,MPI_COMM_WORLD,ierr)
    else
       call mpi_recv(rbuf,j,MPI_REAL, &
                     rr,tag,MPI_COMM_WORLD,status,ierr)
    endif
#endif

    call mpi_barrier(mpi_comm_world,ierr)

    do i=1,npd(2)
      m=(i-1)*6+1
      xvp(:,nploc+i)=rbuf(m:m+5)
    enddo

    do i=1,nploc
      if (xvp(3,i).ge.RLPSS) then
        npu(1)=npu(1)+1
        taglist(npd(1)+npu(1))=i
        sbuf((npu(1)-1)*6+1:npu(1)*6)=xvp(:,i)
      endif
    enddo

    if (npd(1)+npu(1).ge.MAX_TAG) then
      write(*,*) mr,'ran out of tag space'
      call mpi_abort(mpi_comm_world,err_code,ierr)
    endif

    npu(2)=npu(1)
 
    tag = 11
    sr=mod(mr+1,NN)
    rr=mod(mr+NN-1,NN)
    call mpi_sendrecv_replace(npu(2),1,MPI_INTEGER, &
               sr,tag,rr,tag,MPI_COMM_WORLD,status,ierr )

    if (npu(1)*6.ge.MAX_BUFFER) then
      write(*,*) mr,'ran out of buffer space'
      call mpi_abort(mpi_comm_world,err_code,ierr)
    endif

    do i=3,6*npu(1),6
      sbuf(i)=sbuf(i)-RLPSS
      if (sbuf(i).lt.0.0) sbuf(i)=0.0
    enddo

    j=npu(2)*6
    m=npu(1)*6

#ifdef NBPASS
    call mpi_isend(sbuf,m,mpi_real,sr,sr, &
                   mpi_comm_world, srequest, sierr)
    call mpi_irecv(rbuf,j,mpi_real,rr,mr, &
                   mpi_comm_world, rrequest, rierr)
    call mpi_wait(srequest,sstatus,sierr)
    call mpi_wait(rrequest,rstatus,rierr)
#else
    if (mod(mr,2).eq.0) then
      call mpi_send(sbuf,m,MPI_REAL, &
                    sr,tag,MPI_COMM_WORLD,ierr)
    else
      call mpi_recv(rbuf,j,MPI_REAL, &
                    rr,tag,MPI_COMM_WORLD,status,ierr)
    endif

    call mpi_barrier(mpi_comm_world,ierr)

    if (mod(mr,2).eq.1) then
      call mpi_send(sbuf,m,MPI_REAL, &
                    sr,tag,MPI_COMM_WORLD,ierr)
    else
      call mpi_recv(rbuf,j,MPI_REAL, &
                    rr,tag,MPI_COMM_WORLD,status,ierr)
    endif
#endif

    call mpi_barrier(mpi_comm_world,ierr)

    do i=1,npu(2)
      m=(i-1)*6+1
      xvp(:,nploc+npd(2)+i)=rbuf(m:m+5)
    enddo

    if (ngloc.gt.0) call qsorti(taglist(:npu(1)+npd(1))) 

    m=0
    do i=npu(1)+npd(1),1,-1
      xvp(:,taglist(i))=xvp(:,nploc+npu(2)+npd(2)-m)
      m=m+1
    enddo

    do i=0,NN-1
      if (mr.eq.i) write(*,*) 'mr',mr,'nploc',nploc
      call mpi_barrier(mpi_comm_world,ierr)
    enddo

    nploc=nploc+npu(2)+npd(2)-npu(1)-npd(1)

!#ifdef DEBUG
    do i=0,NN-1
      if (mr.eq.i) write(*,*) 'mr',mr,'nploc',nploc
      call mpi_barrier(mpi_comm_world,ierr)
    enddo

    do i=1,nploc
      if (xvp(3,i).lt.0.0.or.xvp(3,i).ge.RLPSS) then
        write(*,*) 'particle out of bounds'
        write(*,*) 'mr',mr,'part num',i,xvp(:,i)
      endif
    enddo 
!#endif

  end subroutine initial_pass
