! combine_proj.f90 - program to combine constituent density projections
! all projection files and projection redshift listing must exist in 
! the local directory.  LPS and NN should correspond to the values used
! in the simulation.

implicit none
integer, parameter :: LPS = 3712
integer, parameter :: NN = 8
integer, parameter :: LPSS = LPS / NN
integer, parameter :: MAX_NUM_RP = 40
real, dimension(LPS,LPS) :: rho_proj_t,rho_proj
real(kind=4) :: z_write,redshift
real(kind=4) :: z_rp(MAX_NUM_RP)
integer :: num_rp, cur_rp,i
character(len=40) :: fn

open(11,file='projections')
do num_rp=1,MAX_NUM_RP
  read(unit=11,err=61,end=71,fmt='(f5.1)') z_rp(num_rp)
enddo
71  num_rp=num_rp-1
61  close(11)
if (num_rp.eq.MAX_NUM_RP) then
  write(*,*) 'too many projections to store'
endif

write(*,*) 'Projections performed at:'
write(*,*) 'z'        
do i=1,num_rp
  write(*,'(f7.4)') z_rp(i)
enddo

do cur_rp=1,num_rp
  z_write=z_rp(cur_rp)

!! First recombine xy plane

  rho_proj=0.
  rho_proj_t=0.

  do i=0,NN-1
    write (fn,'(f7.3,"xy",i1,".proj")') z_write,i 
    open(unit=12,file=fn,form='binary')
    read(12) redshift
    read(12) rho_proj_t
    close(12)
    rho_proj=rho_proj+rho_proj_t
  enddo

  write (fn,'(f7.3,"xy.proj")') z_write 
  open(unit=12,file=fn,form='binary')
  write(12) redshift
  write(12) rho_proj
  close(12)

!! Next recombine xz plane

  rho_proj=0. 
 
  do i=0,NN-1 
    write (fn,'(f7.3,"xz",i1,".proj")') z_write,i 
    open(unit=12,file=fn,form='binary')
    read(12) redshift
    read(12) rho_proj(1:LPS,1+i*LPSS:(i+1)*LPSS)
    close(12)
  enddo

  write (fn,'(f7.3,"xz.proj")') z_write
  open(unit=12,file=fn,form='binary')
  write(12) redshift
  write(12) rho_proj
  close(12)

!! Finally do yz plane

  rho_proj=0. 
 
  do i=0,NN-1 
    write (fn,'(f7.3,"yz",i1,".proj")') z_write,i 
    open(unit=12,file=fn,form='binary')
    read(12) redshift
    read(12) rho_proj(1:LPS,1+i*LPSS:(i+1)*LPSS)
    close(12)
  enddo

  write (fn,'(f7.3,"yz.proj")') z_write
  open(unit=12,file=fn,form='binary')
  write(12) redshift
  write(12) rho_proj
  close(12)
end do

end
