! pair_check.f90 - program to check pairwise force testing
! Hugh Merz - merz@cita.utoronto.ca - Feb 11th 2004
! Output files:
!
! radial_pair.dat contains a line for each particle (2 per pair) with the
! following information:
!
! Magnitude of the radial seperation in fine grid cells
! Magnitude of the total force from the simulation
! Magnitude of the fine mesh force from the simulation
! Magnitude of the exact force
! Fractional force error
! Magnitude of the radial force from the simulation
! Magnitude of the tangential force from the simulation
!
! hist_pair.dat contains a histogram of the number of particles within 
! each percentage of force error

program pair_check
  implicit none

! a computed, e theory
! err = sqrt[ ((ax-ex)^2 + (ay-ey)^2 + (az-ez)^2)/(ex^2+ey^2+ez^2)
!
! rms = sqrt( sum err**2/N  - [sum err/N]**2 )

  integer(kind=4) :: lps,np,i,j,k,np_ri
  real(kind=4) :: mp,dt,r(3),a(3),v(3),F(3),magr,F_sim(3,2)
  real(kind=4) :: magF,magF_sim(2)
  real(kind=4), dimension(12) :: pair,fpair
  real(kind=4) :: F_f(3,2), magF_f(2), rms_err, mean_err, p_err, del_F(2), &
                  frac_err(2),magF_sim_r(2),magF_sim_t(2)
  integer(kind=4) :: hist(100)

  open(unit=34,file='total_pair.dat')
  read(34,'(2i8)') lps, np
  read(34,'(2f20.10)') mp,dt
  open(unit=36,file='fine_pair.dat')
  open(unit=57,file='radial_pair.dat',form='formatted')
  open(unit=58,file='outliers.dat',form='formatted')
  open(unit=48,file='hist_pair.dat',form='formatted')

  hist=0
  np_ri=0
  do i=1,np

    read(34,'(12f20.10)',err=53,end=53) pair(:)
    read(36,'(12f20.10)',err=53,end=53) fpair(:)
    r=0.0
    np_ri=np_ri+1

    do j=1,3
       r(j)=pair(j+6)-pair(j) 
       if (r(j).lt.-real(lps/2,kind=4)) r(j)=r(j)+real(lps,kind=4)
       if (r(j).gt.real(lps/2,kind=4)) r(j)=r(j)-real(lps,kind=4)  
    enddo

    magr=sqrt(r(1)*r(1)+r(2)*r(2)+r(3)*r(3))

    do j=1,3
      a(j)=-mp*r(j)/magr**3.0
      F(j)=a(j)/mp
    enddo

!   F is with G=1 in grid units distance 

    magF=sqrt(F(1)*F(1)+F(2)*F(2)+F(3)*F(3))

    v(:)=a(:)*dt

    do j=1,2 
      F_sim(:,j)=pair(4+(j-1)*6:6+(j-1)*6)/dt/mp
      F_f(:,j)=fpair(4+(j-1)*6:6+(j-1)*6)/dt/mp
    enddo


    do j=1,2
      magF_sim(j)=sqrt(F_sim(1,j)**2+F_sim(2,j)**2+F_sim(3,j)**2)

      magF_sim_r(j)=(F_sim(1,j)*F(1)*(-1.0)**(j) + &
                     F_sim(2,j)*F(2)*(-1.0)**(j) + &
                     F_sim(3,j)*F(3)*(-1.0)**(j))/magF

      magF_sim_t(j)=sqrt(magF_sim(j)**2-magF_sim_r(j)**2)

      magF_f(j)=sqrt(F_f(1,j)**2+F_f(2,j)**2+F_f(3,j)**2)
      del_F(j)=magF-magF_sim(j)
      frac_err(j)=abs(del_F(j))/magF
      mean_err=mean_err+frac_err(j)
      p_err=frac_err(j)*100
      rms_err=rms_err+(frac_err(j))**2
      if (p_err<100) then
        hist(ceiling(p_err))=hist(ceiling(p_err))+1
      else
        hist(100)=hist(100)+1     
      endif 
    enddo

    do j=1,2
      write(57,'(7f16.8)') magr, magF_sim(j), magF_f(j), magF, &
            del_F(j)/magF, magF_sim_r(j), magF_sim_t(j) 
    enddo

    do j=1,2
      if (magr.gt.4.0.and.frac_err(j).gt.0.1) then
        write(58,'(14f16.8)') magr,frac_err(j),pair
        write(58,'(32x,12f16.8)') fpair
      endif
    enddo

  enddo

53 continue

  do i=1,100
    write(48,'(i3,1x,i10)') i,hist(i)
    write(*,*) i,hist(i)
  enddo

  close(48)

  rms_err=sqrt(rms_err/real(2*np_ri))
  mean_err=mean_err/real(2*np_ri)

  write(*,*) 'mean error=',mean_err
  write(*,*) 'rms error=',rms_err

  close(34)
  close(36)
  close(57)
  close(58)

end program pair_check
  
